from __future__ import annotations

import os
from textwrap import dedent
from typing import TYPE_CHECKING
from unittest import mock

import pytest
import requests_mock
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture
from requests import Session

import semantic_release.hvcs.github
from semantic_release.changelog.context import ChangelogMode
from semantic_release.cli.config import ChangelogOutputFormat
from semantic_release.hvcs.github import Github

from tests.const import (
    CHANGELOG_SUBCMD,
    EXAMPLE_HVCS_DOMAIN,
    EXAMPLE_RELEASE_NOTES_TEMPLATE,
    EXAMPLE_REPO_NAME,
    EXAMPLE_REPO_OWNER,
    MAIN_PROG_NAME,
)
from tests.fixtures.example_project import (
    change_to_ex_proj_dir,
    changelog_md_file,
    changelog_rst_file,
    default_md_changelog_insertion_flag,
    default_rst_changelog_insertion_flag,
    example_changelog_md,
    example_changelog_rst,
)
from tests.fixtures.repos import (
    repo_w_git_flow_conventional_commits,
    repo_w_git_flow_emoji_commits,
    repo_w_git_flow_scipy_commits,
    repo_w_git_flow_w_alpha_prereleases_n_conventional_commits,
    repo_w_git_flow_w_alpha_prereleases_n_emoji_commits,
    repo_w_git_flow_w_alpha_prereleases_n_scipy_commits,
    repo_w_git_flow_w_beta_alpha_rev_prereleases_n_conventional_commits,
    repo_w_git_flow_w_beta_alpha_rev_prereleases_n_emoji_commits,
    repo_w_git_flow_w_beta_alpha_rev_prereleases_n_scipy_commits,
    repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits,
    repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits_using_tag_format,
    repo_w_git_flow_w_rc_n_alpha_prereleases_n_emoji_commits,
    repo_w_git_flow_w_rc_n_alpha_prereleases_n_scipy_commits,
    repo_w_github_flow_w_default_release_channel_conventional_commits,
    repo_w_github_flow_w_default_release_channel_emoji_commits,
    repo_w_github_flow_w_default_release_channel_scipy_commits,
    repo_w_github_flow_w_feature_release_channel_conventional_commits,
    repo_w_github_flow_w_feature_release_channel_emoji_commits,
    repo_w_github_flow_w_feature_release_channel_scipy_commits,
    repo_w_no_tags_conventional_commits,
    repo_w_no_tags_emoji_commits,
    repo_w_no_tags_scipy_commits,
    repo_w_trunk_only_conventional_commits,
    repo_w_trunk_only_emoji_commits,
    repo_w_trunk_only_n_prereleases_conventional_commits,
    repo_w_trunk_only_n_prereleases_emoji_commits,
    repo_w_trunk_only_n_prereleases_scipy_commits,
    repo_w_trunk_only_scipy_commits,
)
from tests.util import (
    add_text_to_file,
    assert_exit_code,
    assert_successful_exit_code,
    get_func_qual_name,
    get_release_history_from_context,
)

if TYPE_CHECKING:
    from pathlib import Path
    from typing import TypedDict

    from requests_mock import Mocker

    from semantic_release.commit_parser.conventional.parser import (
        ConventionalCommitParser,
    )
    from semantic_release.commit_parser.emoji import EmojiCommitParser
    from semantic_release.commit_parser.scipy import ScipyCommitParser

    from tests.conftest import RunCliFn
    from tests.e2e.conftest import RetrieveRuntimeContextFn
    from tests.fixtures.example_project import (
        ExProjectDir,
        UpdatePyprojectTomlFn,
        UseReleaseNotesTemplateFn,
    )
    from tests.fixtures.git_repo import (
        BuildRepoFromDefinitionFn,
        BuiltRepoResult,
        CommitConvention,
        CommitDef,
        CommitNReturnChangelogEntryFn,
        GetCommitDefFn,
        GetRepoDefinitionFn,
        GetVersionsFromRepoBuildDefFn,
    )

    class Commit2Section(TypedDict):
        conventional: Commit2SectionCommit
        emoji: Commit2SectionCommit
        scipy: Commit2SectionCommit

    class Commit2SectionCommit(TypedDict):
        commit: CommitDef
        section: str


@pytest.mark.parametrize("arg0", [None, "--post-to-release-tag"])
@pytest.mark.parametrize(
    "repo_result",
    [
        lazy_fixture(repo_fixture)
        for repo_fixture in (
            # Only need to test when it has tags or no tags
            # DO NOT need to consider all repo types as it doesn't change no-op behavior
            repo_w_no_tags_conventional_commits.__name__,
            repo_w_trunk_only_conventional_commits.__name__,
        )
    ],
)
def test_changelog_noop_is_noop(
    repo_result: BuiltRepoResult,
    arg0: str | None,
    run_cli: RunCliFn,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
):
    repo = repo_result["repo"]
    repo_def = repo_result["definition"]
    released_versions = get_versions_from_repo_build_def(repo_def)

    version_str = released_versions[-1] if len(released_versions) > 0 else None

    repo.git.reset("--hard")

    # Set up a requests HTTP session so we can catch the HTTP calls and ensure
    # they're made
    session = Session()
    session.hooks = {"response": [lambda r, *_, **__: r.raise_for_status()]}

    mock_adapter = requests_mock.Adapter()
    mock_adapter.register_uri(
        method=requests_mock.ANY, url=requests_mock.ANY, json={"id": 10001}
    )
    session.mount("http://", mock_adapter)
    session.mount("https://", mock_adapter)

    with mock.patch(
        get_func_qual_name(semantic_release.hvcs.github.build_requests_session),
        return_value=session,
    ), requests_mock.Mocker(session=session) as mocker:
        args = [arg0, f"v{version_str}"] if version_str and arg0 else []
        cli_cmd = [MAIN_PROG_NAME, "--noop", CHANGELOG_SUBCMD, *args]
        result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert not repo.git.status(short=True)
    if args:
        assert not mocker.called
        assert not mock_adapter.called


@pytest.mark.parametrize(
    "changelog_file, insertion_flag",
    [
        (
            # ChangelogOutputFormat.MARKDOWN
            lazy_fixture(example_changelog_md.__name__),
            lazy_fixture(default_md_changelog_insertion_flag.__name__),
        ),
        (
            # ChangelogOutputFormat.RESTRUCTURED_TEXT
            lazy_fixture(example_changelog_rst.__name__),
            lazy_fixture(default_rst_changelog_insertion_flag.__name__),
        ),
    ],
)
@pytest.mark.parametrize(
    "repo_result",
    [
        *[
            lazy_fixture(repo_fixture)
            for repo_fixture in [
                # All commit types and one without a release
                repo_w_no_tags_conventional_commits.__name__,
                repo_w_trunk_only_conventional_commits.__name__,
                repo_w_trunk_only_emoji_commits.__name__,
                repo_w_trunk_only_scipy_commits.__name__,
            ]
        ],
        *[
            pytest.param(lazy_fixture(repo_fixture), marks=pytest.mark.comprehensive)
            for repo_fixture in [
                # repo_w_no_tags_conventional_commits.__name__,
                repo_w_no_tags_emoji_commits.__name__,
                repo_w_no_tags_scipy_commits.__name__,
                # repo_w_trunk_only_conventional_commits.__name__,
                # repo_w_trunk_only_emoji_commits.__name__,
                # repo_w_trunk_only_scipy_commits.__name__,
                repo_w_trunk_only_n_prereleases_conventional_commits.__name__,
                repo_w_trunk_only_n_prereleases_emoji_commits.__name__,
                repo_w_trunk_only_n_prereleases_scipy_commits.__name__,
                repo_w_github_flow_w_default_release_channel_conventional_commits.__name__,
                repo_w_github_flow_w_default_release_channel_emoji_commits.__name__,
                repo_w_github_flow_w_default_release_channel_scipy_commits.__name__,
                repo_w_github_flow_w_feature_release_channel_conventional_commits.__name__,
                repo_w_github_flow_w_feature_release_channel_emoji_commits.__name__,
                repo_w_github_flow_w_feature_release_channel_scipy_commits.__name__,
                repo_w_git_flow_conventional_commits.__name__,
                repo_w_git_flow_emoji_commits.__name__,
                repo_w_git_flow_scipy_commits.__name__,
                repo_w_git_flow_w_beta_alpha_rev_prereleases_n_conventional_commits.__name__,
                repo_w_git_flow_w_beta_alpha_rev_prereleases_n_emoji_commits.__name__,
                repo_w_git_flow_w_beta_alpha_rev_prereleases_n_scipy_commits.__name__,
                repo_w_git_flow_w_alpha_prereleases_n_conventional_commits.__name__,
                repo_w_git_flow_w_alpha_prereleases_n_emoji_commits.__name__,
                repo_w_git_flow_w_alpha_prereleases_n_scipy_commits.__name__,
                repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits.__name__,
                repo_w_git_flow_w_rc_n_alpha_prereleases_n_emoji_commits.__name__,
                repo_w_git_flow_w_rc_n_alpha_prereleases_n_scipy_commits.__name__,
                repo_w_git_flow_w_rc_n_alpha_prereleases_n_conventional_commits_using_tag_format.__name__,
            ]
        ],
    ],
)
def test_changelog_content_regenerated(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    changelog_file: Path,
    insertion_flag: str,
):
    # Set the project configurations
    update_pyproject_toml(
        "tool.semantic_release.changelog.mode", ChangelogMode.INIT.value
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.default_templates.changelog_file",
        str(changelog_file.name),
    )

    # Because we are in init mode, the insertion flag is not present in the changelog
    # we must take it out manually because our repo generation fixture includes it automatically
    with changelog_file.open(newline=os.linesep) as rfd:
        # use os.linesep here because the insertion flag is os-specific
        # but convert the content to universal newlines for comparison
        expected_changelog_content = (
            rfd.read().replace(f"{insertion_flag}{os.linesep}", "").replace("\r", "")
        )

    # Remove the changelog and then check that we can regenerate it
    os.remove(str(changelog_file.resolve()))

    # Act
    cli_cmd = [MAIN_PROG_NAME, CHANGELOG_SUBCMD]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Check that the changelog file was re-created
    assert changelog_file.exists()

    actual_content = changelog_file.read_text()

    # Check that the changelog content is the same as before
    assert expected_changelog_content == actual_content


@pytest.mark.parametrize(
    "changelog_file, insertion_flag",
    [
        (
            # ChangelogOutputFormat.MARKDOWN
            lazy_fixture(example_changelog_md.__name__),
            lazy_fixture(default_md_changelog_insertion_flag.__name__),
        ),
        (
            # ChangelogOutputFormat.RESTRUCTURED_TEXT
            lazy_fixture(example_changelog_rst.__name__),
            lazy_fixture(default_rst_changelog_insertion_flag.__name__),
        ),
    ],
)
@pytest.mark.usefixtures(change_to_ex_proj_dir.__name__)
def test_changelog_content_regenerated_masked_initial_release(
    build_repo_from_definition: BuildRepoFromDefinitionFn,
    get_repo_definition_4_trunk_only_repo_w_tags: GetRepoDefinitionFn,
    example_project_dir: ExProjectDir,
    run_cli: RunCliFn,
    changelog_file: Path,
    insertion_flag: str,
):
    build_definition = get_repo_definition_4_trunk_only_repo_w_tags(
        commit_type="conventional",
        mask_initial_release=True,
        extra_configs={
            "tool.semantic_release.changelog.default_templates.changelog_file": str(
                changelog_file.name
            ),
            "tool.semantic_release.changelog.mode": ChangelogMode.INIT.value,
        },
    )
    build_repo_from_definition(example_project_dir, build_definition)

    # Because we are in init mode, the insertion flag is not present in the changelog
    # we must take it out manually because our repo generation fixture includes it automatically
    with changelog_file.open(newline=os.linesep) as rfd:
        # use os.linesep here because the insertion flag is os-specific
        # but convert the content to universal newlines for comparison
        expected_changelog_content = (
            rfd.read().replace(f"{insertion_flag}{os.linesep}", "").replace("\r", "")
        )

    # Remove the changelog and then check that we can regenerate it
    os.remove(str(changelog_file.resolve()))

    # Act
    cli_cmd = [MAIN_PROG_NAME, CHANGELOG_SUBCMD]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Check that the changelog file was re-created
    assert changelog_file.exists()

    actual_content = changelog_file.read_text()

    # Check that the changelog content is the same as before
    assert expected_changelog_content == actual_content


@pytest.mark.parametrize(
    "changelog_file",
    [
        lazy_fixture(example_changelog_md.__name__),
        lazy_fixture(example_changelog_rst.__name__),
    ],
)
@pytest.mark.parametrize(
    "repo_result",
    [
        lazy_fixture(repo_fixture)
        for repo_fixture in [
            repo_w_trunk_only_conventional_commits.__name__,
            repo_w_trunk_only_emoji_commits.__name__,
            repo_w_trunk_only_scipy_commits.__name__,
        ]
    ],
)
def test_changelog_update_mode_unchanged(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    changelog_file: Path,
):
    """
    Given that the changelog file already exists for the current release,
    When the changelog command is run in "update" mode,
    Then the changelog file is not modified.
    """
    # Set the project configurations
    update_pyproject_toml(
        "tool.semantic_release.changelog.mode", ChangelogMode.UPDATE.value
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.default_templates.changelog_file",
        str(changelog_file.name),
    )

    # Capture the expected changelog content
    expected_changelog_content = changelog_file.read_text()

    # Act
    cli_cmd = [MAIN_PROG_NAME, CHANGELOG_SUBCMD]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Check that the changelog file was re-created
    assert changelog_file.exists()

    actual_content = changelog_file.read_text()

    # Check that the changelog content is the same as before
    assert expected_changelog_content == actual_content


@pytest.mark.parametrize(
    "changelog_file",
    [
        lazy_fixture(example_changelog_md.__name__),
        lazy_fixture(example_changelog_rst.__name__),
    ],
)
@pytest.mark.parametrize(
    "repo_result",
    [
        lazy_fixture(repo_fixture)
        for repo_fixture in [
            repo_w_no_tags_conventional_commits.__name__,
            repo_w_no_tags_emoji_commits.__name__,
            repo_w_no_tags_scipy_commits.__name__,
            repo_w_trunk_only_conventional_commits.__name__,
            repo_w_trunk_only_emoji_commits.__name__,
            repo_w_trunk_only_scipy_commits.__name__,
        ]
    ],
)
def test_changelog_update_mode_no_prev_changelog(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    changelog_file: Path,
):
    """
    Given that the changelog file does not exist,
    When the changelog command is run in "update" mode,
    Then the changelog file is initialized with the default content.
    """
    # Set the project configurations
    update_pyproject_toml(
        "tool.semantic_release.changelog.mode", ChangelogMode.UPDATE.value
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.default_templates.changelog_file",
        str(changelog_file.name),
    )

    # Capture the expected changelog content
    expected_changelog_content = changelog_file.read_text()

    # Remove any previous changelog to update
    os.remove(str(changelog_file.resolve()))

    # Act
    cli_cmd = [MAIN_PROG_NAME, CHANGELOG_SUBCMD]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Check that the changelog file was re-created
    assert changelog_file.exists()

    actual_content = changelog_file.read_text()

    # Check that the changelog content is the same as before
    assert expected_changelog_content == actual_content


@pytest.mark.parametrize(
    "changelog_file, insertion_flag",
    [
        (
            # ChangelogOutputFormat.MARKDOWN
            lazy_fixture(example_changelog_md.__name__),
            lazy_fixture(default_md_changelog_insertion_flag.__name__),
        ),
        (
            # ChangelogOutputFormat.RESTRUCTURED_TEXT
            lazy_fixture(example_changelog_rst.__name__),
            lazy_fixture(default_rst_changelog_insertion_flag.__name__),
        ),
    ],
)
@pytest.mark.parametrize(
    "repo_result",
    [
        lazy_fixture(repo_fixture)
        for repo_fixture in [
            repo_w_trunk_only_conventional_commits.__name__,
            repo_w_trunk_only_emoji_commits.__name__,
            repo_w_trunk_only_scipy_commits.__name__,
        ]
    ],
)
def test_changelog_update_mode_no_flag(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    changelog_file: Path,
    insertion_flag: str,
):
    """
    Given a changelog template without the insertion flag,
    When the changelog command is run in "update" mode,
    Then the changelog is not modified.
    """
    # Set the project configurations
    update_pyproject_toml(
        "tool.semantic_release.changelog.mode", ChangelogMode.UPDATE.value
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.default_templates.changelog_file",
        str(changelog_file.name),
    )

    # Remove the insertion flag from the changelog
    changelog_file.write_text(
        changelog_file.read_text().replace(
            f"{insertion_flag}\n",
            "",
            1,
        )
    )

    # Capture the expected changelog content
    expected_changelog_content = changelog_file.read_text()

    # Act
    cli_cmd = [MAIN_PROG_NAME, CHANGELOG_SUBCMD]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Ensure changelog exists
    assert changelog_file.exists()

    actual_content = changelog_file.read_text()

    # Check that the changelog content is the same as before
    assert expected_changelog_content == actual_content


@pytest.mark.parametrize(
    "changelog_format, changelog_file",
    [
        (
            ChangelogOutputFormat.MARKDOWN,
            lazy_fixture(changelog_md_file.__name__),
        ),
        (
            ChangelogOutputFormat.RESTRUCTURED_TEXT,
            lazy_fixture(changelog_rst_file.__name__),
        ),
    ],
)
@pytest.mark.parametrize(
    "repo_result",
    [
        lazy_fixture(repo_fixture)
        for repo_fixture in [
            # MUST HAVE at least 2 tags!
            repo_w_trunk_only_conventional_commits.__name__,
            repo_w_trunk_only_emoji_commits.__name__,
            repo_w_trunk_only_scipy_commits.__name__,
        ]
    ],
)
def test_changelog_update_mode_no_header(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    changelog_format: ChangelogOutputFormat,
    changelog_file: Path,
    default_md_changelog_insertion_flag: str,
    default_rst_changelog_insertion_flag: str,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
):
    """
    Given a changelog template with the insertion flag at the beginning of the file,
    When the changelog command is run in "update" mode,
    Then the changelog is rebuilt with the latest release prepended to the existing content.
    """
    repo = repo_result["repo"]

    # Mappings of correct fixtures to use based on the changelog format
    insertion_flags = {
        ChangelogOutputFormat.MARKDOWN: (
            "# CHANGELOG{ls}{ls}{flag}".format(
                ls=os.linesep,
                flag=default_md_changelog_insertion_flag,
            )
        ),
        ChangelogOutputFormat.RESTRUCTURED_TEXT: (
            ".. _changelog:{ls}{ls}{h1_border}{ls}CHANGELOG{ls}{h1_border}{ls}{ls}{flag}".format(
                ls=os.linesep,
                h1_border="=" * 9,
                flag=default_rst_changelog_insertion_flag,
            )
        ),
    }

    # Select the correct insertion flag based on the format
    insertion_flag = insertion_flags[changelog_format]

    # Set the project configurations
    update_pyproject_toml(
        "tool.semantic_release.changelog.mode", ChangelogMode.UPDATE.value
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.default_templates.changelog_file",
        str(changelog_file.name),
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.insertion_flag",
        insertion_flag,
    )

    # Capture the expected changelog content of current release
    with changelog_file.open(newline=os.linesep) as rfd:
        expected_changelog_content = rfd.read()

    # Reset changelog file to last release
    previous_tag = f'v{get_versions_from_repo_build_def(repo_result["definition"])[-2]}'
    repo.git.checkout(previous_tag, "--", str(changelog_file.name))

    # Act
    cli_cmd = [MAIN_PROG_NAME, CHANGELOG_SUBCMD]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Ensure changelog exists
    assert changelog_file.exists()

    with changelog_file.open(newline=os.linesep) as rfd:
        actual_content = rfd.read()

    # Check that the changelog content is the same as before
    assert expected_changelog_content == actual_content


@pytest.mark.parametrize(
    "changelog_format, changelog_file, insertion_flag",
    [
        (
            ChangelogOutputFormat.MARKDOWN,
            lazy_fixture(example_changelog_md.__name__),
            lazy_fixture(default_md_changelog_insertion_flag.__name__),
        ),
        (
            ChangelogOutputFormat.RESTRUCTURED_TEXT,
            lazy_fixture(example_changelog_rst.__name__),
            lazy_fixture(default_rst_changelog_insertion_flag.__name__),
        ),
    ],
)
@pytest.mark.parametrize(
    "repo_result",
    [
        lazy_fixture(repo_fixture)
        for repo_fixture in [
            # MUST HAVE at least 2 tags!
            repo_w_trunk_only_conventional_commits.__name__,
            repo_w_trunk_only_emoji_commits.__name__,
            repo_w_trunk_only_scipy_commits.__name__,
        ]
    ],
)
def test_changelog_update_mode_no_footer(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    changelog_format: ChangelogOutputFormat,
    changelog_file: Path,
    insertion_flag: str,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
):
    """
    Given a changelog template with the insertion flag at the end of the file,
    When the changelog command is run in "update" mode,
    Then the changelog is rebuilt with only the latest release.
    """
    repo_result["repo"]

    # Mappings of correct fixtures to use based on the changelog format
    prev_version_tag = (
        f"v{get_versions_from_repo_build_def(repo_result['definition'])[-2]}"
    )
    split_flags = {
        ChangelogOutputFormat.MARKDOWN: f"\n\n## {prev_version_tag}",
        ChangelogOutputFormat.RESTRUCTURED_TEXT: f"\n\n.. _changelog-{prev_version_tag}:",
    }

    # Select the correct variable based on the format
    split_flag = split_flags[changelog_format]

    # Set the project configurations
    update_pyproject_toml(
        "tool.semantic_release.changelog.mode", ChangelogMode.UPDATE.value
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.default_templates.changelog_file",
        str(changelog_file.name),
    )

    # Capture the expected changelog content of current release (w/ universal newlines)
    # NOTE: universal newlines is fine because we use our split flag above is also universal
    expected_changelog_content = changelog_file.read_text().split(split_flag)[0]

    # Determine the contents to save while truncating the rest
    with changelog_file.open(newline=os.linesep) as rfd:
        # read file contents grabbing only the text before the insertion flag
        truncated_contents = str.join(
            "",
            [
                rfd.read().split(insertion_flag)[0],
                insertion_flag,
                os.linesep,
            ],
        )

    # Remove any text after the insertion flag
    # force output to not perform any newline translations
    with changelog_file.open(mode="w", newline="") as wfd:
        # overwrite the file with truncated contents
        wfd.write(truncated_contents)
        wfd.flush()

    # Act
    cli_cmd = [MAIN_PROG_NAME, CHANGELOG_SUBCMD]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Ensure changelog exists
    assert changelog_file.exists()

    actual_content = changelog_file.read_text()

    # Check that the changelog content only includes the latest release as there
    # is no previous release information as the insertion flag is at the end of the file
    assert expected_changelog_content == actual_content


@pytest.mark.parametrize(
    "changelog_file, insertion_flag",
    [
        (
            # ChangelogOutputFormat.MARKDOWN
            lazy_fixture(example_changelog_md.__name__),
            lazy_fixture(default_md_changelog_insertion_flag.__name__),
        ),
        (
            # ChangelogOutputFormat.RESTRUCTURED_TEXT
            lazy_fixture(example_changelog_rst.__name__),
            lazy_fixture(default_rst_changelog_insertion_flag.__name__),
        ),
    ],
)
@pytest.mark.parametrize(
    "repo_result",
    [
        lazy_fixture(repo_fixture)
        for repo_fixture in [
            # Must not have a single release/tag
            repo_w_no_tags_conventional_commits.__name__,
            repo_w_no_tags_emoji_commits.__name__,
            repo_w_no_tags_scipy_commits.__name__,
        ]
    ],
)
def test_changelog_update_mode_no_releases(
    repo_result: BuiltRepoResult,
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    changelog_file: Path,
    insertion_flag: str,
):
    """
    Given the repository has no releases and the user has provided a initialized changelog,
    When the changelog command is run in "update" mode,
    Then the changelog is populated with unreleased changes.
    """
    # Set the project configurations
    update_pyproject_toml(
        "tool.semantic_release.changelog.mode", ChangelogMode.UPDATE.value
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.default_templates.changelog_file",
        str(changelog_file.name),
    )

    # Custom text to maintain (must be different from the default)
    custom_text = "---{ls}{ls}Custom footer text{ls}".format(ls=os.linesep)

    # Capture and modify the current changelog content to become the expected output
    # We much use os.linesep here since the insertion flag is os-specific
    with changelog_file.open(newline=os.linesep) as rfd:
        initial_changelog_parts = rfd.read().split(insertion_flag)

    # content is os-specific because of the insertion flag & how we read the original file
    expected_changelog_content = str.join(
        insertion_flag,
        [
            initial_changelog_parts[0],
            str.join(
                os.linesep,
                [
                    initial_changelog_parts[1],
                    "",
                    custom_text,
                ],
            ),
        ],
    )

    # Grab the Unreleased changelog & create the initalized user changelog
    # force output to not perform any newline translations
    with changelog_file.open(mode="w", newline="") as wfd:
        wfd.write(
            str.join(
                insertion_flag,
                [initial_changelog_parts[0], f"{os.linesep * 2}{custom_text}"],
            )
        )

    # Act
    cli_cmd = [MAIN_PROG_NAME, CHANGELOG_SUBCMD]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Ensure changelog exists
    assert changelog_file.exists()

    # Capture the new changelog content (os aware because of expected content)
    with changelog_file.open(newline=os.linesep) as rfd:
        actual_content = rfd.read()

    # Check that the changelog footer is maintained and updated with Unreleased info
    assert expected_changelog_content == actual_content


@pytest.mark.parametrize(
    "changelog_format, changelog_file, insertion_flag",
    [
        (
            ChangelogOutputFormat.MARKDOWN,
            lazy_fixture(example_changelog_md.__name__),
            lazy_fixture(default_md_changelog_insertion_flag.__name__),
        ),
        (
            ChangelogOutputFormat.RESTRUCTURED_TEXT,
            lazy_fixture(example_changelog_rst.__name__),
            lazy_fixture(default_rst_changelog_insertion_flag.__name__),
        ),
    ],
)
@pytest.mark.parametrize(
    "repo_result, commit_type",
    [
        (lazy_fixture(repo_fixture), repo_fixture.split("_")[-2])
        for repo_fixture in [
            repo_w_trunk_only_conventional_commits.__name__,
            repo_w_trunk_only_emoji_commits.__name__,
            repo_w_trunk_only_scipy_commits.__name__,
        ]
    ],
)
def test_changelog_update_mode_unreleased_n_released(
    repo_result: BuiltRepoResult,
    commit_type: CommitConvention,
    changelog_format: ChangelogOutputFormat,
    run_cli: RunCliFn,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    example_git_ssh_url: str,
    file_in_repo: str,
    commit_n_rtn_changelog_entry: CommitNReturnChangelogEntryFn,
    changelog_file: Path,
    insertion_flag: str,
    get_commit_def_of_conventional_commit: GetCommitDefFn[ConventionalCommitParser],
    get_commit_def_of_emoji_commit: GetCommitDefFn[EmojiCommitParser],
    get_commit_def_of_scipy_commit: GetCommitDefFn[ScipyCommitParser],
    default_conventional_parser: ConventionalCommitParser,
    default_emoji_parser: EmojiCommitParser,
    default_scipy_parser: ScipyCommitParser,
):
    """
    Given there are unreleased changes and a previous release in the changelog,
    When the changelog command is run in "update" mode,
    Then the changelog is only updated with the unreleased changes.
    """
    repo = repo_result["repo"]

    # Set the project configurations
    update_pyproject_toml(
        "tool.semantic_release.changelog.mode", ChangelogMode.UPDATE.value
    )
    update_pyproject_toml(
        "tool.semantic_release.changelog.default_templates.changelog_file",
        str(changelog_file.name),
    )

    commit_n_section: Commit2Section = {
        "conventional": {
            "commit": get_commit_def_of_conventional_commit(
                "perf: improve the performance of the application",
                parser=default_conventional_parser,
            ),
            "section": "Performance Improvements",
        },
        "emoji": {
            "commit": get_commit_def_of_emoji_commit(
                ":zap: improve the performance of the application",
                parser=default_emoji_parser,
            ),
            "section": ":zap:",
        },
        "scipy": {
            "commit": get_commit_def_of_scipy_commit(
                "MAINT: fix an issue",
                parser=default_scipy_parser,
            ),
            "section": "Fix",
        },
    }

    # Custom text to maintain (must be different from the default)
    custom_text = "---\n\nCustom footer text\n"

    # Update the changelog with the custom footer text
    changelog_file.write_text(
        str.join(
            "\n\n",
            [
                changelog_file.read_text(),
                custom_text,
            ],
        )
    )

    # Capture the current changelog content so we can estimate the expected output
    # We much use os.linesep here since the insertion flag is os-specific
    with changelog_file.open(newline=os.linesep) as rfd:
        initial_changelog_parts = rfd.read().split(insertion_flag)

    # Make a change to the repo to create unreleased changes
    add_text_to_file(repo, file_in_repo)
    unreleased_commit_entry = commit_n_rtn_changelog_entry(
        repo,
        commit_n_section[commit_type]["commit"],
    )

    with mock.patch.dict(os.environ, {}, clear=True):
        hvcs = Github(example_git_ssh_url, hvcs_domain=EXAMPLE_HVCS_DOMAIN)
        assert hvcs.repo_name  # force caching of repo values (ignoring the env)

    unreleased_change_variants = {
        ChangelogOutputFormat.MARKDOWN: dedent(
            f"""
            ## Unreleased

            ### {commit_n_section[commit_type]["section"]}

            - {unreleased_commit_entry['desc'].capitalize()}
              ([`{unreleased_commit_entry['sha'][:7]}`]({hvcs.commit_hash_url(unreleased_commit_entry['sha'])}))
            """
        ),
        ChangelogOutputFormat.RESTRUCTURED_TEXT: dedent(
            f"""
            .. _changelog-unreleased:

            Unreleased
            ==========

            {commit_n_section[commit_type]["section"]}
            {"-" * len(commit_n_section[commit_type]["section"])}

            * {unreleased_commit_entry['desc'].capitalize()} (`{unreleased_commit_entry['sha'][:7]}`_)

            .. _{unreleased_commit_entry['sha'][:7]}: {hvcs.commit_hash_url(unreleased_commit_entry['sha'])}
            """
        ),
    }

    # Normalize line endings to the OS-specific line ending
    unreleased_changes = str.join(
        os.linesep,
        [
            line.replace("\r", "")
            for line in unreleased_change_variants[changelog_format].split("\n")
        ],
    )

    # Generate the expected changelog content (os aware because of insertion flag & initial parts)
    expected_changelog_content = str.join(
        insertion_flag,
        [
            initial_changelog_parts[0],
            str.join(
                "",
                [
                    os.linesep,
                    # Unreleased changes
                    unreleased_changes,
                    # Previous release notes
                    initial_changelog_parts[1],
                ],
            ),
        ],
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, CHANGELOG_SUBCMD]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Ensure changelog exists
    assert changelog_file.exists()

    # Capture the new changelog content (os aware because of expected content)
    with changelog_file.open(newline=os.linesep) as rfd:
        actual_content = rfd.read()

    # Check that the changelog content is the same as before
    assert expected_changelog_content == actual_content


# Just need to test that it works for "a" project, not all
@pytest.mark.usefixtures(repo_w_trunk_only_n_prereleases_conventional_commits.__name__)
@pytest.mark.parametrize(
    "args", [("--post-to-release-tag", "v1.99.91910000000000000000000000000")]
)
def test_changelog_release_tag_not_in_history(
    args: list[str],
    run_cli: RunCliFn,
):
    # Act
    cli_cmd = [MAIN_PROG_NAME, CHANGELOG_SUBCMD, *args]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_exit_code(2, result, cli_cmd)
    assert "not in release history" in result.stderr.lower()


@pytest.mark.usefixtures(repo_w_trunk_only_n_prereleases_conventional_commits.__name__)
@pytest.mark.parametrize(
    "args",
    [
        ("--post-to-release-tag", "v0.1.0"),  #      first release
        ("--post-to-release-tag", "v0.1.1-rc.1"),  # second release
        ("--post-to-release-tag", "v0.2.0"),  #      latest release
    ],
)
def test_changelog_post_to_release(args: list[str], run_cli: RunCliFn):
    # Set up a requests HTTP session so we can catch the HTTP calls and ensure they're
    # made

    session = Session()
    session.hooks = {"response": [lambda r, *_, **__: r.raise_for_status()]}

    mock_adapter = requests_mock.Adapter()
    mock_adapter.register_uri(
        method=requests_mock.ANY, url=requests_mock.ANY, json={"id": 10001}
    )
    session.mount("http://", mock_adapter)
    session.mount("https://", mock_adapter)

    expected_request_url = "{api_url}/repos/{owner}/{repo_name}/releases".format(
        api_url=f"https://{EXAMPLE_HVCS_DOMAIN}/api/v3",  # GitHub API URL
        owner=EXAMPLE_REPO_OWNER,
        repo_name=EXAMPLE_REPO_NAME,
    )

    # Patch out env vars that affect changelog URLs but only get set in e.g.
    # Github actions
    with mock.patch(
        # Patching the specific module's reference to the build_requests_session function
        f"{semantic_release.hvcs.github.__name__}.{semantic_release.hvcs.github.build_requests_session.__name__}",
        return_value=session,
    ) as build_requests_session_mock:
        # Act
        cli_cmd = [MAIN_PROG_NAME, CHANGELOG_SUBCMD, *args]
        result = run_cli(
            cli_cmd[1:],
            env={
                "CI": "true",
                "VIRTUAL_ENV": os.getenv("VIRTUAL_ENV", "./.venv"),
            },
        )

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)
    assert build_requests_session_mock.called
    assert mock_adapter.called
    assert mock_adapter.last_request is not None
    assert expected_request_url == mock_adapter.last_request.url


@pytest.mark.parametrize(
    "repo_result",
    [lazy_fixture(repo_w_trunk_only_n_prereleases_conventional_commits.__name__)],
)
def test_custom_release_notes_template(
    repo_result: BuiltRepoResult,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    use_release_notes_template: UseReleaseNotesTemplateFn,
    retrieve_runtime_context: RetrieveRuntimeContextFn,
    post_mocker: Mocker,
    run_cli: RunCliFn,
) -> None:
    """Verify the template `.release_notes.md.j2` from `template_dir` is used."""
    expected_call_count = 1
    version = get_versions_from_repo_build_def(repo_result["definition"])[-1]

    # Setup
    use_release_notes_template()
    runtime_context = retrieve_runtime_context(repo_result["repo"])
    release_history = get_release_history_from_context(runtime_context)
    release = release_history.released[version]
    tag = runtime_context.version_translator.str_to_tag(str(version))

    expected_release_notes = (
        runtime_context.template_environment.from_string(EXAMPLE_RELEASE_NOTES_TEMPLATE)
        .render(release=release)
        .rstrip()
        + os.linesep
    )

    # ensure normalized line endings after render
    expected_release_notes = str.join(
        os.linesep,
        str.split(expected_release_notes.replace("\r", ""), "\n"),
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, CHANGELOG_SUBCMD, "--post-to-release-tag", tag]
    result = run_cli(cli_cmd[1:])

    # Assert
    assert_successful_exit_code(result, cli_cmd)
    assert expected_call_count == post_mocker.call_count
    assert post_mocker.last_request is not None

    actual_notes = post_mocker.last_request.json()["body"]
    assert expected_release_notes == actual_notes


@pytest.mark.usefixtures(repo_w_trunk_only_conventional_commits.__name__)
def test_changelog_default_on_empty_template_dir(
    example_changelog_md: Path,
    changelog_template_dir: Path,
    example_project_template_dir: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    run_cli: RunCliFn,
):
    # Setup: Make sure default changelog doesn't already exist
    example_changelog_md.unlink(missing_ok=True)

    # Setup: Create an empty template directory
    example_project_template_dir.mkdir(parents=True, exist_ok=True)

    # Setup: Set the templates directory in the configuration
    update_pyproject_toml(
        "tool.semantic_release.changelog.template_dir",
        str(changelog_template_dir),
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, CHANGELOG_SUBCMD]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Check that our default changelog was created because the user's template dir was empty
    assert example_changelog_md.exists()


@pytest.mark.usefixtures(repo_w_trunk_only_conventional_commits.__name__)
def test_changelog_default_on_incorrect_config_template_file(
    example_changelog_md: Path,
    changelog_template_dir: Path,
    example_project_template_dir: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
    run_cli: RunCliFn,
):
    # Setup: Make sure default changelog doesn't already exist
    example_changelog_md.unlink(missing_ok=True)

    # Setup: Create a file of the same name as the template directory
    example_project_template_dir.parent.mkdir(parents=True, exist_ok=True)
    example_project_template_dir.touch()

    # Setup: Set the templates directory as the file in the configuration
    update_pyproject_toml(
        "tool.semantic_release.changelog.template_dir",
        str(changelog_template_dir),
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, CHANGELOG_SUBCMD]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_successful_exit_code(result, cli_cmd)

    # Check that our default changelog was created because the user's template dir was empty
    assert example_changelog_md.exists()


@pytest.mark.parametrize("bad_changelog_file_str", ("/etc/passwd", "../../.ssh/id_rsa"))
@pytest.mark.usefixtures(repo_w_trunk_only_conventional_commits.__name__)
def test_changelog_prevent_malicious_path_traversal_file(
    update_pyproject_toml: UpdatePyprojectTomlFn,
    bad_changelog_file_str: str,
    run_cli: RunCliFn,
):
    # Setup: A malicious path traversal filepath outside of the repository
    update_pyproject_toml(
        "tool.semantic_release.changelog.changelog_file",
        bad_changelog_file_str,
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, "--noop", CHANGELOG_SUBCMD]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_exit_code(1, result, cli_cmd)
    assert (
        "Changelog file destination must be inside of the repository directory."
        in result.stderr
    )


@pytest.mark.parametrize("template_dir_path", ("~/.ssh", "../../.ssh"))
@pytest.mark.usefixtures(repo_w_trunk_only_conventional_commits.__name__)
def test_changelog_prevent_external_path_traversal_dir(
    update_pyproject_toml: UpdatePyprojectTomlFn,
    template_dir_path: str,
    run_cli: RunCliFn,
):
    # Setup: A malicious path traversal filepath outside of the repository
    update_pyproject_toml(
        "tool.semantic_release.changelog.template_dir",
        template_dir_path,
    )

    # Act
    cli_cmd = [MAIN_PROG_NAME, "--noop", CHANGELOG_SUBCMD]
    result = run_cli(cli_cmd[1:])

    # Evaluate
    assert_exit_code(1, result, cli_cmd)
    assert (
        "Template directory must be inside of the repository directory."
        in result.stderr
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from semantic_release.hvcs._base import HvcsBase

from tests.const import EXAMPLE_REPO_NAME, EXAMPLE_REPO_OWNER
from tests.fixtures.git_repo import example_git_https_url, example_git_ssh_url

if TYPE_CHECKING:
    from typing import Any, Callable


class ArbitraryHvcs(HvcsBase):
    def remote_url(self, use_token: bool) -> str:
        return super().remote_url(use_token)

    def get_changelog_context_filters(self) -> tuple[Callable[..., Any], ...]:
        return super().get_changelog_context_filters()


@pytest.mark.parametrize(
    "remote_url, repo_name",
    [
        (lazy_fixture(example_git_ssh_url.__name__), EXAMPLE_REPO_NAME),
        (lazy_fixture(example_git_https_url.__name__), EXAMPLE_REPO_NAME),
        ("git@my.corp.custom.domain:very_serious/business.git", "business"),
    ],
)
def test_get_repository_owner(remote_url, repo_name):
    client = ArbitraryHvcs(remote_url)
    assert client.repo_name == repo_name


@pytest.mark.parametrize(
    "remote_url, owner",
    [
        (lazy_fixture(example_git_ssh_url.__name__), EXAMPLE_REPO_OWNER),
        (lazy_fixture(example_git_https_url.__name__), EXAMPLE_REPO_OWNER),
        ("git@my.corp.custom.domain:very_serious/business.git", "very_serious"),
    ],
)
def test_get_repository_name(remote_url, owner):
    client = ArbitraryHvcs(remote_url)
    assert client.owner == owner


@pytest.mark.parametrize(
    "bad_url",
    [
        "a" * 25,
        "https://a/b/c/d/.git",
        "https://github.com/wrong",
        "git@gitlab.com/somewhere",
    ],
)
def test_hvcs_parse_error(bad_url: str):
    with pytest.raises(ValueError):
        ArbitraryHvcs(bad_url)



from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, cast

import tomlkit
from deprecated.sphinx import deprecated
from dotty_dict import Dotty

from semantic_release.cli.util import noop_report
from semantic_release.globals import logger
from semantic_release.version.declarations.enum import VersionStampType
from semantic_release.version.declarations.i_version_replacer import IVersionReplacer
from semantic_release.version.version import Version


class TomlVersionDeclaration(IVersionReplacer):
    def __init__(
        self, path: Path | str, search_text: str, stamp_format: VersionStampType
    ) -> None:
        self._content: str | None = None
        self._path = Path(path).resolve()
        self._stamp_format = stamp_format
        self._search_text = search_text

    @property
    def content(self) -> str:
        """A cached property that stores the content of the configured source file."""
        if self._content is None:
            logger.debug("No content stored, reading from source file %s", self._path)

            if not self._path.exists():
                raise FileNotFoundError(f"path {self._path!r} does not exist")

            self._content = self._path.read_text()

        return self._content

    @content.deleter
    def content(self) -> None:
        self._content = None

    @deprecated(
        version="9.20.0",
        reason="Function is unused and will be removed in a future release",
    )
    def parse(self) -> set[Version]:  # pragma: no cover
        """Look for the version in the source content"""
        content = self._load()
        maybe_version: str = content.get(self._search_text)  # type: ignore[return-value]
        if maybe_version is not None:
            logger.debug(
                "Found a key %r that looks like a version (%r)",
                self._search_text,
                maybe_version,
            )
            valid_version = Version.parse(maybe_version)
            return {valid_version} if valid_version else set()
        # Maybe in future raise error if not found?
        return set()

    def replace(self, new_version: Version) -> str:
        """
        Replace the version in the source content with `new_version`, and return the
        updated content.
        """
        content = self._load()
        if self._search_text in content:
            logger.info(
                "found %r in source file contents, replacing with %s",
                self._search_text,
                new_version,
            )
            content[self._search_text] = (
                new_version.as_tag()
                if self._stamp_format == VersionStampType.TAG_FORMAT
                else str(new_version)
            )

        return tomlkit.dumps(cast(Dict[str, Any], content))

    def _load(self) -> Dotty:
        """Load the content of the source file into a Dotty for easier searching"""
        return Dotty(tomlkit.loads(self.content))

    def update_file_w_version(
        self, new_version: Version, noop: bool = False
    ) -> Path | None:
        if noop:
            if not self._path.exists():
                noop_report(
                    f"FILE NOT FOUND: cannot stamp version in non-existent file {self._path!r}",
                )
                return None

            if self._search_text not in self._load():
                noop_report(
                    f"VERSION PATTERN NOT FOUND: no version to stamp in file {self._path!r}",
                )
                return None

            return self._path

        new_content = self.replace(new_version)
        if new_content == self.content:
            return None

        self._path.write_text(new_content)
        del self.content

        return self._path

    @classmethod
    def from_string_definition(cls, replacement_def: str) -> TomlVersionDeclaration:
        """
        create an instance of self from a string representing one item
        of the "version_toml" list in the configuration
        """
        parts = replacement_def.split(":", maxsplit=2)

        if len(parts) <= 1:
            raise ValueError(
                f"Invalid TOML replacement definition {replacement_def!r}, missing ':'"
            )

        if len(parts) == 2:
            # apply default version_type of "number_format" (ie. "1.2.3")
            parts = [*parts, VersionStampType.NUMBER_FORMAT.value]

        path, search_text, version_type = parts

        try:
            stamp_type = VersionStampType(version_type)
        except ValueError as err:
            raise ValueError(
                str.join(
                    " ",
                    [
                        "Invalid stamp type, must be one of:",
                        str.join(", ", [e.value for e in VersionStampType]),
                    ],
                )
            ) from err

        return cls(path, search_text, stamp_type)



"""Semantic Release Global Variables."""

from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING

from semantic_release.enums import SemanticReleaseLogLevels

if TYPE_CHECKING:
    from logging import Logger

# GLOBAL VARIABLES
log_level: SemanticReleaseLogLevels = SemanticReleaseLogLevels.WARNING
"""int: Logging level for semantic-release"""

logger: Logger = getLogger(__package__)
"""Logger for semantic-release"""



from __future__ import annotations

from pathlib import Path
from re import compile as regexp
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from semantic_release.version.declarations.enum import VersionStampType
from semantic_release.version.declarations.i_version_replacer import IVersionReplacer
from semantic_release.version.declarations.toml import TomlVersionDeclaration
from semantic_release.version.version import Version

from tests.fixtures.git_repo import default_tag_format_str

if TYPE_CHECKING:
    from re import Pattern


def test_toml_declaration_is_version_replacer():
    """
    Given the class TomlVersionDeclaration or an instance of it,
    When the class is evaluated as a subclass or an instance of,
    Then the evaluation is true
    """
    assert issubclass(TomlVersionDeclaration, IVersionReplacer)

    toml_instance = TomlVersionDeclaration(
        "file", "project.version", VersionStampType.NUMBER_FORMAT
    )
    assert isinstance(toml_instance, IVersionReplacer)


@pytest.mark.parametrize(
    str.join(
        ", ",
        [
            "replacement_def",
            "tag_format",
            "starting_contents",
            "resulting_contents",
            "next_version",
            "test_file",
        ],
    ),
    [
        pytest.param(
            replacement_def,
            tag_format,
            starting_contents,
            resulting_contents,
            next_version,
            test_file,
            id=test_id,
        )
        for test_file in ["test_file.toml"]
        for next_version in ["1.2.3"]
        for test_id, replacement_def, tag_format, starting_contents, resulting_contents in [
            (
                "Default number format for project.version",
                f"{test_file}:project.version",
                # irrelevant for this case
                lazy_fixture(default_tag_format_str.__name__),
                # Uses equals separator with single quotes
                dedent(
                    """\
                    [project]
                    version = '1.0.0'
                    """
                ),
                dedent(
                    f"""\
                    [project]
                    version = "{next_version}"
                    """
                ),
            ),
            (
                "Explicit number format for project.version",
                f"{test_file}:project.version:{VersionStampType.NUMBER_FORMAT.value}",
                # irrelevant for this case
                lazy_fixture(default_tag_format_str.__name__),
                # Uses equals separator with double quotes
                dedent(
                    """\
                    [project]
                    version = "1.0.0"
                    """
                ),
                dedent(
                    f"""\
                    [project]
                    version = "{next_version}"
                    """
                ),
            ),
            (
                "Using default tag format for toml string variable",
                f"{test_file}:version:{VersionStampType.TAG_FORMAT.value}",
                lazy_fixture(default_tag_format_str.__name__),
                # Uses equals separator with single quotes
                '''version = "v1.0.0"''',
                f'''version = "v{next_version}"''',
            ),
            (
                "Using custom tag format for toml string variable",
                f"{test_file}:version:{VersionStampType.TAG_FORMAT.value}",
                "module-v{version}",
                # Uses equals separator with double quotes
                '''version = "module-v1.0.0"''',
                f'''version = "module-v{next_version}"''',
            ),
        ]
    ],
)
def test_toml_declaration_from_definition(
    replacement_def: str,
    tag_format: str,
    starting_contents: str,
    resulting_contents: str,
    next_version: str,
    test_file: str,
    change_to_ex_proj_dir: None,
):
    """
    Given a file with a formatted version string,
    When update_file_w_version() is called with a new version,
    Then the file is updated with the new version string in the specified tag or number format

    Version variables can be separated by either "=", ":", "@", or ':=' with optional whitespace
    between operator and variable name. The variable name or values can also be wrapped in either
    single or double quotes.
    """
    # Setup: create file with initial contents
    expected_filepath = Path(test_file).resolve()
    expected_filepath.write_text(starting_contents)

    # Create Pattern Replacer
    version_replacer = TomlVersionDeclaration.from_string_definition(replacement_def)

    # Act: apply version change
    actual_file_modified = version_replacer.update_file_w_version(
        new_version=Version.parse(next_version, tag_format=tag_format),
        noop=False,
    )

    # Evaluate
    actual_contents = Path(test_file).read_text()
    assert resulting_contents == actual_contents
    assert expected_filepath == actual_file_modified


def test_toml_declaration_no_file_change(
    change_to_ex_proj_dir: None,
):
    """
    Given a configured stamp file is already up-to-date,
    When update_file_w_version() is called with the same version,
    Then the file is not modified and no path is returned
    """
    test_file = "test_file"
    next_version = Version.parse("1.2.3")
    starting_contents = dedent(
        f"""\
        [project]
        version = "{next_version}"
        """
    )

    # Setup: create file with initial contents
    Path(test_file).write_text(starting_contents)

    # Create Pattern Replacer
    version_replacer = TomlVersionDeclaration.from_string_definition(
        f"{test_file}:project.version:{VersionStampType.NUMBER_FORMAT.value}",
    )

    # Act: apply version change
    file_modified = version_replacer.update_file_w_version(
        new_version=next_version,
        noop=False,
    )

    # Evaluate
    actual_contents = Path(test_file).read_text()
    assert starting_contents == actual_contents
    assert file_modified is None


def test_toml_declaration_error_on_missing_file():
    # Initialization should not fail or do anything intensive
    version_replacer = TomlVersionDeclaration.from_string_definition(
        "nonexistent_file:version",
    )

    with pytest.raises(FileNotFoundError):
        version_replacer.update_file_w_version(
            new_version=Version.parse("1.2.3"),
            noop=False,
        )


def test_toml_declaration_no_version_in_file(
    change_to_ex_proj_dir: None,
):
    test_file = "test_file"
    expected_filepath = Path(test_file).resolve()
    starting_contents = dedent(
        """\
        [project]
        name = "example"
        """
    )

    # Setup: create file with initial contents
    expected_filepath.write_text(starting_contents)

    # Create Pattern Replacer
    version_replacer = TomlVersionDeclaration.from_string_definition(
        f"{test_file}:project.version:{VersionStampType.NUMBER_FORMAT.value}",
    )

    file_modified = version_replacer.update_file_w_version(
        new_version=Version.parse("1.2.3"),
        noop=False,
    )

    # Evaluate
    actual_contents = expected_filepath.read_text()
    assert file_modified is None
    assert starting_contents == actual_contents


def test_toml_declaration_noop_is_noop(
    change_to_ex_proj_dir: None,
):
    test_file = "test_file"
    expected_filepath = Path(test_file).resolve()
    starting_contents = dedent(
        """\
        [project]
        version = '1.0.0'
        """
    )

    # Setup: create file with initial contents
    expected_filepath.write_text(starting_contents)

    # Create Pattern Replacer
    version_replacer = TomlVersionDeclaration.from_string_definition(
        f"{test_file}:project.version:{VersionStampType.NUMBER_FORMAT.value}",
    )

    # Act: apply version change
    file_modified = version_replacer.update_file_w_version(
        new_version=Version.parse("1.2.3"),
        noop=True,
    )

    # Evaluate
    actual_contents = Path(test_file).read_text()
    assert starting_contents == actual_contents
    assert expected_filepath == file_modified


def test_toml_declaration_noop_warning_on_missing_file(
    capsys: pytest.CaptureFixture[str],
):
    version_replacer = TomlVersionDeclaration.from_string_definition(
        "nonexistent_file:version",
    )

    file_to_modify = version_replacer.update_file_w_version(
        new_version=Version.parse("1.2.3"),
        noop=True,
    )

    # Evaluate
    assert file_to_modify is None
    assert (
        "FILE NOT FOUND: cannot stamp version in non-existent file"
        in capsys.readouterr().err
    )


def test_toml_declaration_noop_warning_on_no_version_in_file(
    capsys: pytest.CaptureFixture[str],
    change_to_ex_proj_dir: None,
):
    test_file = "test_file"
    starting_contents = dedent(
        """\
        [project]
        name = "example"
        """
    )

    # Setup: create file with initial contents
    Path(test_file).write_text(starting_contents)

    # Create Pattern Replacer
    version_replacer = TomlVersionDeclaration.from_string_definition(
        f"{test_file}:project.version:{VersionStampType.NUMBER_FORMAT.value}",
    )

    file_to_modify = version_replacer.update_file_w_version(
        new_version=Version.parse("1.2.3"),
        noop=True,
    )

    # Evaluate
    assert file_to_modify is None
    assert (
        "VERSION PATTERN NOT FOUND: no version to stamp in file"
        in capsys.readouterr().err
    )


@pytest.mark.parametrize(
    "replacement_def, error_msg",
    [
        pytest.param(
            replacement_def,
            error_msg,
            id=str(error_msg),
        )
        for replacement_def, error_msg in [
            (
                f"{Path(__file__)!s}",
                regexp(r"Invalid TOML replacement definition .*, missing ':'"),
            ),
            (
                f"{Path(__file__)!s}:tool.poetry.version:not_a_valid_version_type",
                "Invalid stamp type, must be one of:",
            ),
        ]
    ],
)
def test_toml_declaration_w_invalid_definition(
    replacement_def: str,
    error_msg: Pattern[str] | str,
):
    """
    check if TomlVersionDeclaration raises ValueError when loaded
    from invalid strings given in the config file
    """
    with pytest.raises(ValueError, match=error_msg):
        TomlVersionDeclaration.from_string_definition(replacement_def)



from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

import click
from click.core import ParameterSource
from git import InvalidGitRepositoryError
from pydantic import ValidationError

from semantic_release.cli.config import (
    RawConfig,
    RuntimeContext,
)
from semantic_release.cli.util import load_raw_config_file, rprint
from semantic_release.errors import (
    DetachedHeadGitError,
    InvalidConfiguration,
    NotAReleaseBranch,
)

if TYPE_CHECKING:  # pragma: no cover
    from semantic_release.cli.config import GlobalCommandLineOptions

    class CliContext(click.Context):
        obj: CliContextObj


class CliContextObj:
    def __init__(
        self,
        ctx: click.Context,
        logger: logging.Logger,
        global_opts: GlobalCommandLineOptions,
    ) -> None:
        self.ctx = ctx
        self.logger = logger
        self.global_opts = global_opts
        self._raw_config: RawConfig | None = None
        self._runtime_ctx: RuntimeContext | None = None

    @property
    def raw_config(self) -> RawConfig:
        if self._raw_config is None:
            self._raw_config = self._init_raw_config()
        return self._raw_config

    @property
    def runtime_ctx(self) -> RuntimeContext:
        """
        Lazy load the runtime context. This is done to avoid configuration loading when
        the command is not run. This is useful for commands like `--help` and `--version`
        """
        if self._runtime_ctx is None:
            self._runtime_ctx = self._init_runtime_ctx()
        return self._runtime_ctx

    def _init_raw_config(self) -> RawConfig:
        config_path = Path(self.global_opts.config_file)
        conf_file_exists = config_path.exists()
        was_conf_file_user_provided = bool(
            self.ctx.get_parameter_source("config_file")
            not in (
                ParameterSource.DEFAULT,
                ParameterSource.DEFAULT_MAP,
            )
        )

        # TODO: Evaluate Exeception catches
        try:
            if was_conf_file_user_provided and not conf_file_exists:
                raise FileNotFoundError(  # noqa: TRY301
                    f"File {self.global_opts.config_file} does not exist"
                )

            config_obj = (
                {} if not conf_file_exists else load_raw_config_file(config_path)
            )
            if not config_obj:
                self.logger.info(
                    "configuration empty, falling back to default configuration"
                )

            return RawConfig.model_validate(config_obj)
        except FileNotFoundError as exc:
            click.echo(str(exc), err=True)
            self.ctx.exit(2)
        except (
            ValidationError,
            InvalidConfiguration,
            InvalidGitRepositoryError,
        ) as exc:
            click.echo(str(exc), err=True)
            self.ctx.exit(1)

    def _init_runtime_ctx(self) -> RuntimeContext:
        # TODO: Evaluate Exception catches
        try:
            runtime = RuntimeContext.from_raw_config(
                self.raw_config,
                global_cli_options=self.global_opts,
            )
        except NotAReleaseBranch as exc:
            rprint(f"[bold {'red' if self.global_opts.strict else 'orange1'}]{exc!s}")
            # If not strict, exit 0 so other processes can continue. For example, in
            # multibranch CI it might be desirable to run a non-release branch's pipeline
            # without specifying conditional execution of PSR based on branch name
            self.ctx.exit(2 if self.global_opts.strict else 0)
        except (
            DetachedHeadGitError,
            InvalidConfiguration,
            InvalidGitRepositoryError,
            ValidationError,
        ) as exc:
            click.echo(str(exc), err=True)
            self.ctx.exit(1)

        # This allows us to mask secrets in the logging
        # by applying it to all the configured handlers
        for handler in logging.getLogger().handlers:
            handler.addFilter(runtime.masker)

        return runtime



"""Tests for version command with shallow repositories."""

from __future__ import annotations

from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING, cast

import pytest
from git import Repo
from pytest_lazy_fixtures.lazy_fixture import lf as lazy_fixture

from semantic_release.hvcs.github import Github

from tests.const import MAIN_PROG_NAME, VERSION_SUBCMD
from tests.fixtures.example_project import change_to_ex_proj_dir
from tests.fixtures.repos import repo_w_trunk_only_conventional_commits
from tests.fixtures.repos.trunk_based_dev.repo_w_tags import (
    build_trunk_only_repo_w_tags,
)
from tests.util import assert_successful_exit_code, temporary_working_directory

if TYPE_CHECKING:
    from requests_mock import Mocker

    from tests.conftest import RunCliFn
    from tests.fixtures.example_project import ExProjectDir, UpdatePyprojectTomlFn
    from tests.fixtures.git_repo import (
        BuildSpecificRepoFn,
        CommitConvention,
        GetCfgValueFromDefFn,
        GetGitRepo4DirFn,
        GetVersionsFromRepoBuildDefFn,
    )


@pytest.mark.parametrize(
    "repo_fixture_name, build_repo_fn",
    [
        (
            repo_fixture_name,
            lazy_fixture(build_repo_fn_name),
        )
        for repo_fixture_name, build_repo_fn_name in [
            (
                repo_w_trunk_only_conventional_commits.__name__,
                build_trunk_only_repo_w_tags.__name__,
            ),
        ]
    ],
)
@pytest.mark.usefixtures(change_to_ex_proj_dir.__name__)
def test_version_w_shallow_repo_unshallows(
    repo_fixture_name: str,
    run_cli: RunCliFn,
    build_repo_fn: BuildSpecificRepoFn,
    example_project_dir: ExProjectDir,
    git_repo_for_directory: GetGitRepo4DirFn,
    post_mocker: Mocker,
    get_cfg_value_from_def: GetCfgValueFromDefFn,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    pyproject_toml_file: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
) -> None:
    """
    Test that the version command automatically unshallows a shallow repository.

    Given a shallow repository,
    When running the version command,
    Then the repository should be unshallowed and release should succeed
    """
    remote_name = "origin"

    # Create a bare remote (simulating origin)
    local_origin = Repo.init(str(example_project_dir / "local_origin"), bare=True)

    # build target repo into a temporary directory
    target_repo_dir = example_project_dir / repo_fixture_name
    commit_type: CommitConvention = (
        repo_fixture_name.split("commits", 1)[0].split("_")[-2]  # type: ignore[assignment]
    )
    target_repo_definition = build_repo_fn(
        repo_name=repo_fixture_name,
        commit_type=commit_type,
        dest_dir=target_repo_dir,
    )
    target_git_repo = git_repo_for_directory(target_repo_dir)

    # Configure the source repo to use the bare remote (removing any existing 'origin')
    with suppress(AttributeError):
        target_git_repo.delete_remote(target_git_repo.remotes[remote_name])

    target_git_repo.create_remote(remote_name, str(local_origin.working_dir))

    # Remove last release before pushing to upstream
    tag_format_str = cast(
        "str", get_cfg_value_from_def(target_repo_definition, "tag_format_str")
    )
    latest_tag = tag_format_str.format(
        version=get_versions_from_repo_build_def(target_repo_definition)[-1]
    )
    target_git_repo.git.tag("-d", latest_tag)
    target_git_repo.git.reset("--hard", "HEAD~1")

    # TODO: when available, switch this to use hvcs=none or similar config to avoid token use for push
    update_pyproject_toml(
        "tool.semantic_release.remote.ignore_token_for_push",
        True,
        target_repo_dir / pyproject_toml_file,
    )
    target_git_repo.git.commit(amend=True, no_edit=True, all=True)

    # push the current state to establish the remote (cannot push tags and branches at the same time)
    target_git_repo.git.push(remote_name, all=True)  # all branches
    target_git_repo.git.push(remote_name, tags=True)  # all tags

    # ensure bare remote HEAD points to the active branch so clones can checkout
    local_origin.git.symbolic_ref(
        "HEAD", f"refs/heads/{target_git_repo.active_branch.name}"
    )

    # current remote tags
    remote_origin_tags_before = {tag.name for tag in local_origin.tags}

    # Create a shallow clone from the remote using file:// protocol for depth support
    shallow_repo = Repo.clone_from(
        f"file://{local_origin.working_dir}",
        str(example_project_dir / "shallow_clone"),
        no_local=True,
        depth=1,
    )
    with shallow_repo.config_writer("repository") as config:
        config.set_value("core", "hookspath", "")
        config.set_value("commit", "gpgsign", False)
        config.set_value("tag", "gpgsign", False)

    with shallow_repo:
        # Verify it's a shallow clone
        shallow_file = Path(shallow_repo.git_dir, "shallow")
        assert shallow_file.exists(), "Repository should be shallow"

    # Capture expected values from the full repo
    expected_vcs_url_post = 1
    commit_sha_before = shallow_repo.head.commit.hexsha

    # Run PSR on the shallow clone
    with temporary_working_directory(str(shallow_repo.working_dir)):
        cli_cmd = [MAIN_PROG_NAME, VERSION_SUBCMD, "--patch"]
        result = run_cli(cli_cmd[1:], env={Github.DEFAULT_ENV_TOKEN_NAME: "1234"})

    # Initial execution check
    assert_successful_exit_code(result, cli_cmd)

    # Take measurements after running PSR
    remote_origin_tags_after = {tag.name for tag in local_origin.tags}
    different_tags = remote_origin_tags_after.difference(remote_origin_tags_before)
    with shallow_repo:
        parent_commit_shas = [
            parent.hexsha for parent in shallow_repo.head.commit.parents
        ]
        commit_sha_after = shallow_repo.head.commit.hexsha

    # Verify the shallow file is gone (repo was unshallowed)
    assert not shallow_file.exists(), "Repository should be unshallowed"

    # Verify release was successful
    assert commit_sha_before != commit_sha_after, "Expected commit SHA to change"
    assert (
        commit_sha_before in parent_commit_shas
    ), "Expected new commit to be created on HEAD"
    assert (
        latest_tag in different_tags
    ), "Expected a new tag to be created and pushed to remote"
    assert expected_vcs_url_post == post_mocker.call_count  # 1x vcs release created


@pytest.mark.parametrize(
    "repo_fixture_name, build_repo_fn",
    [
        (
            repo_fixture_name,
            lazy_fixture(build_repo_fn_name),
        )
        for repo_fixture_name, build_repo_fn_name in [
            (
                repo_w_trunk_only_conventional_commits.__name__,
                build_trunk_only_repo_w_tags.__name__,
            ),
        ]
    ],
)
@pytest.mark.usefixtures(change_to_ex_proj_dir.__name__)
def test_version_noop_w_shallow_repo(
    repo_fixture_name: str,
    run_cli: RunCliFn,
    build_repo_fn: BuildSpecificRepoFn,
    example_project_dir: ExProjectDir,
    git_repo_for_directory: GetGitRepo4DirFn,
    post_mocker: Mocker,
    get_cfg_value_from_def: GetCfgValueFromDefFn,
    get_versions_from_repo_build_def: GetVersionsFromRepoBuildDefFn,
    pyproject_toml_file: Path,
    update_pyproject_toml: UpdatePyprojectTomlFn,
) -> None:
    """
    Test that the version command in noop mode reports unshallow action.

    Given a shallow repository,
    When running the version command with --noop,
    Then the command should report what it would do but not actually unshallow
    """
    remote_name = "origin"

    # Create a bare remote (simulating origin)
    local_origin = Repo.init(str(example_project_dir / "local_origin"), bare=True)

    # build target repo into a temporary directory
    target_repo_dir = example_project_dir / repo_fixture_name
    commit_type: CommitConvention = (
        repo_fixture_name.split("commits", 1)[0].split("_")[-2]  # type: ignore[assignment]
    )
    target_repo_definition = build_repo_fn(
        repo_name=repo_fixture_name,
        commit_type=commit_type,
        dest_dir=target_repo_dir,
    )
    target_git_repo = git_repo_for_directory(target_repo_dir)

    # Configure the source repo to use the bare remote (removing any existing 'origin')
    with suppress(AttributeError):
        target_git_repo.delete_remote(target_git_repo.remotes[remote_name])

    target_git_repo.create_remote(remote_name, str(local_origin.working_dir))

    # Remove last release before pushing to upstream
    tag_format_str = cast(
        "str", get_cfg_value_from_def(target_repo_definition, "tag_format_str")
    )
    latest_tag = tag_format_str.format(
        version=get_versions_from_repo_build_def(target_repo_definition)[-1]
    )
    target_git_repo.git.tag("-d", latest_tag)
    target_git_repo.git.reset("--hard", "HEAD~1")

    # TODO: when available, switch this to use hvcs=none or similar config to avoid token use for push
    update_pyproject_toml(
        "tool.semantic_release.remote.ignore_token_for_push",
        True,
        target_repo_dir / pyproject_toml_file,
    )
    target_git_repo.git.commit(amend=True, no_edit=True, all=True)

    # push the current state to establish the remote (cannot push tags and branches at the same time)
    target_git_repo.git.push(remote_name, all=True)  # all branches
    target_git_repo.git.push(remote_name, tags=True)  # all tags

    # ensure bare remote HEAD points to the active branch so clones can checkout
    local_origin.git.symbolic_ref(
        "HEAD", f"refs/heads/{target_git_repo.active_branch.name}"
    )

    # Create a shallow clone from the remote using file:// protocol for depth support
    shallow_repo = Repo.clone_from(
        f"file://{local_origin.working_dir}",
        str(example_project_dir / "shallow_clone"),
        no_local=True,
        depth=1,
    )
    with shallow_repo.config_writer("repository") as config:
        config.set_value("core", "hookspath", "")
        config.set_value("commit", "gpgsign", False)
        config.set_value("tag", "gpgsign", False)

    with shallow_repo:
        # Verify it's a shallow clone
        shallow_file = Path(shallow_repo.git_dir, "shallow")
        assert shallow_file.exists(), "Repository should be shallow"

    # Capture expected values from the full repo
    expected_vcs_url_post = 0
    commit_sha_before = shallow_repo.head.commit.hexsha
    remote_origin_tags_before = {tag.name for tag in local_origin.tags}

    # Run PSR in noop mode on the shallow clone
    with temporary_working_directory(str(shallow_repo.working_dir)):
        cli_cmd = [MAIN_PROG_NAME, "--noop", VERSION_SUBCMD, "--patch"]
        result = run_cli(cli_cmd[1:], env={Github.DEFAULT_ENV_TOKEN_NAME: "1234"})

    # Initial execution check
    assert_successful_exit_code(result, cli_cmd)

    # Take measurements after running PSR
    remote_origin_tags_after = {tag.name for tag in local_origin.tags}
    different_tags = remote_origin_tags_after.difference(remote_origin_tags_before)
    with shallow_repo:
        commit_sha_after = shallow_repo.head.commit.hexsha

    # Verify the shallow file still exists (repo was NOT actually unshallowed in noop)
    assert shallow_file.exists(), "Repository should still be shallow in noop mode"

    # Verify no actual changes were made
    assert (
        commit_sha_before == commit_sha_after
    ), "Expected commit SHA to remain unchanged in noop mode"
    assert not different_tags, "Expected no new tags to be created in noop mode"
    assert expected_vcs_url_post == post_mocker.call_count



from __future__ import annotations

import fnmatch
import glob
import os
import re
from typing import TYPE_CHECKING
from unittest import mock
from urllib.parse import urlencode

import pytest
import requests_mock
from requests import HTTPError, Response, Session
from requests.auth import _basic_auth_str

from semantic_release.hvcs.gitea import Gitea
from semantic_release.hvcs.token_auth import TokenAuth

from tests.const import (
    EXAMPLE_HVCS_DOMAIN,
    EXAMPLE_REPO_NAME,
    EXAMPLE_REPO_OWNER,
    RELEASE_NOTES,
)
from tests.fixtures.example_project import init_example_project

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Generator

    from tests.conftest import NetrcFileFn


@pytest.fixture
def default_gitea_client() -> Generator[Gitea, None, None]:
    remote_url = (
        f"git@{Gitea.DEFAULT_DOMAIN}:{EXAMPLE_REPO_OWNER}/{EXAMPLE_REPO_NAME}.git"
    )
    with mock.patch.dict(os.environ, {}, clear=True):
        yield Gitea(remote_url=remote_url)


@pytest.mark.parametrize(
    str.join(
        ", ",
        [
            "patched_os_environ",
            "hvcs_domain",
            "expected_hvcs_domain",
            "insecure",
        ],
    ),
    # NOTE: Gitea does not have a different api domain
    [
        # Default values
        ({}, None, f"https://{Gitea.DEFAULT_DOMAIN}", False),
        (
            # Gather domain from environment
            {"GITEA_SERVER_URL": "https://special.custom.server/"},
            None,
            "https://special.custom.server",
            False,
        ),
        (
            # Custom domain with path prefix (derives from environment)
            {"GITEA_SERVER_URL": "https://special.custom.server/vcs/"},
            None,
            "https://special.custom.server/vcs",
            False,
        ),
        (
            # Ignore environment & use provided parameter value (ie from user config)
            {"GITEA_SERVER_URL": "https://special.custom.server/"},
            f"https://{EXAMPLE_HVCS_DOMAIN}",
            f"https://{EXAMPLE_HVCS_DOMAIN}",
            False,
        ),
        (
            # Allow insecure http connections explicitly
            {},
            f"http://{EXAMPLE_HVCS_DOMAIN}",
            f"http://{EXAMPLE_HVCS_DOMAIN}",
            True,
        ),
        (
            # Infer insecure connection from user configuration
            {},
            EXAMPLE_HVCS_DOMAIN,
            f"http://{EXAMPLE_HVCS_DOMAIN}",
            True,
        ),
    ],
)
@pytest.mark.parametrize(
    "remote_url",
    [
        f"git@{Gitea.DEFAULT_DOMAIN}:{EXAMPLE_REPO_OWNER}/{EXAMPLE_REPO_NAME}.git",
        f"https://{Gitea.DEFAULT_DOMAIN}/{EXAMPLE_REPO_OWNER}/{EXAMPLE_REPO_NAME}.git",
    ],
)
@pytest.mark.parametrize("token", ("abc123", None))
def test_gitea_client_init(
    patched_os_environ: dict[str, str],
    hvcs_domain: str | None,
    expected_hvcs_domain: str,
    remote_url: str,
    token: str | None,
    insecure: bool,
):
    with mock.patch.dict(os.environ, patched_os_environ, clear=True):
        client = Gitea(
            remote_url=remote_url,
            hvcs_domain=hvcs_domain,
            token=token,
            allow_insecure=insecure,
        )

        # Evaluate (expected -> actual)
        assert expected_hvcs_domain == client.hvcs_domain.url
        assert f"{expected_hvcs_domain}/api/v1" == str(client.api_url)
        assert token == client.token
        assert remote_url == client._remote_url
        assert hasattr(client, "session")
        assert isinstance(getattr(client, "session", None), Session)


@pytest.mark.parametrize(
    "hvcs_domain, insecure",
    [
        (f"ftp://{EXAMPLE_HVCS_DOMAIN}", False),
        (f"ftp://{EXAMPLE_HVCS_DOMAIN}", True),
        (f"http://{EXAMPLE_HVCS_DOMAIN}", False),
    ],
)
def test_gitea_client_init_with_invalid_scheme(hvcs_domain: str, insecure: bool):
    with pytest.raises(ValueError), mock.patch.dict(os.environ, {}, clear=True):
        Gitea(
            remote_url=f"https://{EXAMPLE_HVCS_DOMAIN}/{EXAMPLE_REPO_OWNER}/{EXAMPLE_REPO_NAME}.git",
            hvcs_domain=hvcs_domain,
            allow_insecure=insecure,
        )


def test_gitea_get_repository_owner_and_name(default_gitea_client: Gitea):
    expected_result = (EXAMPLE_REPO_OWNER, EXAMPLE_REPO_NAME)

    # Execute method under test
    result = default_gitea_client._get_repository_owner_and_name()

    # Evaluate (expected -> actual)
    assert expected_result == result


@pytest.mark.parametrize(
    "use_token, token, remote_url, expected_auth_url",
    [
        (
            False,
            "",
            f"git@{Gitea.DEFAULT_DOMAIN}:custom/example.git",
            f"git@{Gitea.DEFAULT_DOMAIN}:custom/example.git",
        ),
        (
            True,
            "",
            f"git@{Gitea.DEFAULT_DOMAIN}:custom/example.git",
            f"git@{Gitea.DEFAULT_DOMAIN}:custom/example.git",
        ),
        (
            False,
            "aabbcc",
            f"git@{Gitea.DEFAULT_DOMAIN}:custom/example.git",
            f"git@{Gitea.DEFAULT_DOMAIN}:custom/example.git",
        ),
        (
            True,
            "aabbcc",
            f"git@{Gitea.DEFAULT_DOMAIN}:custom/example.git",
            f"https://aabbcc@{Gitea.DEFAULT_DOMAIN}/custom/example.git",
        ),
    ],
)
def test_remote_url(
    default_gitea_client: Gitea,
    use_token: bool,
    token: str,
    remote_url: str,
    expected_auth_url: str,
):
    default_gitea_client._remote_url = remote_url
    default_gitea_client.token = token
    assert expected_auth_url == default_gitea_client.remote_url(use_token=use_token)


def test_commit_hash_url(default_gitea_client: Gitea):
    sha = "hashashash"
    expected_url = "{server}/{owner}/{repo}/commit/{sha}".format(
        server=default_gitea_client.hvcs_domain.url,
        owner=default_gitea_client.owner,
        repo=default_gitea_client.repo_name,
        sha=sha,
    )
    assert expected_url == default_gitea_client.commit_hash_url(sha)


def test_commit_hash_url_w_custom_server():
    """
    Test the commit hash URL generation for a self-hosted Bitbucket server with prefix.

    ref: https://github.com/python-semantic-release/python-semantic-release/issues/1204
    """
    sha = "244f7e11bcb1e1ce097db61594056bc2a32189a0"
    expected_url = "{server}/{owner}/{repo}/commit/{sha}".format(
        server=f"https://{EXAMPLE_HVCS_DOMAIN}/projects/demo-foo",
        owner="foo",
        repo=EXAMPLE_REPO_NAME,
        sha=sha,
    )

    with mock.patch.dict(os.environ, {}, clear=True):
        actual_url = Gitea(
            remote_url=f"https://{EXAMPLE_HVCS_DOMAIN}/projects/demo-foo/foo/{EXAMPLE_REPO_NAME}.git",
            hvcs_domain=f"https://{EXAMPLE_HVCS_DOMAIN}/projects/demo-foo",
        ).commit_hash_url(sha)

    assert expected_url == actual_url


@pytest.mark.parametrize("issue_number", (666, "666", "#666"))
def test_issue_url(default_gitea_client: Gitea, issue_number: int | str):
    expected_url = "{server}/{owner}/{repo}/issues/{issue_number}".format(
        server=default_gitea_client.hvcs_domain.url,
        owner=default_gitea_client.owner,
        repo=default_gitea_client.repo_name,
        issue_number=str(issue_number).lstrip("#"),
    )
    assert expected_url == default_gitea_client.issue_url(issue_num=issue_number)


@pytest.mark.parametrize("pr_number", (666, "666", "#666"))
def test_pull_request_url(default_gitea_client: Gitea, pr_number: int | str):
    expected_url = "{server}/{owner}/{repo}/pulls/{pr_number}".format(
        server=default_gitea_client.hvcs_domain.url,
        owner=default_gitea_client.owner,
        repo=default_gitea_client.repo_name,
        pr_number=str(pr_number).lstrip("#"),
    )
    actual_url = default_gitea_client.pull_request_url(pr_number=pr_number)
    assert expected_url == actual_url


@pytest.mark.parametrize("release_id", (42, 666))
def test_asset_upload_url(default_gitea_client: Gitea, release_id: int):
    expected_url = "{server}/repos/{owner}/{repo}/releases/{release_id}/assets".format(
        server=default_gitea_client.api_url,
        owner=default_gitea_client.owner,
        repo=default_gitea_client.repo_name,
        release_id=release_id,
    )
    actual_url = default_gitea_client.asset_upload_url(release_id=release_id)
    assert expected_url == actual_url


############
# Tests which need http response mocking
############


gitea_matcher = re.compile(rf"^https://{Gitea.DEFAULT_DOMAIN}")
gitea_api_matcher = re.compile(
    rf"^https://{Gitea.DEFAULT_DOMAIN}{Gitea.DEFAULT_API_PATH}"
)


@pytest.mark.parametrize("status_code", [201])
@pytest.mark.parametrize("mock_release_id", range(3))
@pytest.mark.parametrize("prerelease", (True, False))
def test_create_release_succeeds(
    default_gitea_client: Gitea,
    mock_release_id: int,
    prerelease: bool,
    status_code: int,
):
    tag = "v1.0.0"
    expected_num_requests = 1
    expected_http_method = "POST"
    expected_request_url = "{api_url}/repos/{owner}/{repo_name}/releases".format(
        api_url=default_gitea_client.api_url,
        owner=default_gitea_client.owner,
        repo_name=default_gitea_client.repo_name,
    )
    expected_request_body = {
        "tag_name": tag,
        "name": tag,
        "body": RELEASE_NOTES,
        "draft": False,
        "prerelease": prerelease,
    }

    with requests_mock.Mocker(session=default_gitea_client.session) as m:
        # mock the response
        m.register_uri(
            "POST",
            gitea_api_matcher,
            json={"id": mock_release_id},
            status_code=status_code,
        )

        # Execute method under test
        actual_rtn_val = default_gitea_client.create_release(
            tag, RELEASE_NOTES, prerelease
        )

        # Evaluate (expected -> actual)
        assert mock_release_id == actual_rtn_val
        assert m.called
        assert expected_num_requests == len(m.request_history)
        assert expected_http_method == m.last_request.method
        assert expected_request_url == m.last_request.url
        assert expected_request_body == m.last_request.json()


@pytest.mark.parametrize("status_code", (400, 409))
@pytest.mark.parametrize("mock_release_id", range(3))
@pytest.mark.parametrize("prerelease", (True, False))
def test_create_release_fails(
    default_gitea_client: Gitea,
    mock_release_id: int,
    prerelease: bool,
    status_code: int,
):
    tag = "v1.0.0"
    expected_num_requests = 1
    expected_http_method = "POST"
    expected_request_url = "{api_url}/repos/{owner}/{repo_name}/releases".format(
        api_url=default_gitea_client.api_url,
        owner=default_gitea_client.owner,
        repo_name=default_gitea_client.repo_name,
    )
    expected_request_body = {
        "tag_name": tag,
        "name": tag,
        "body": RELEASE_NOTES,
        "draft": False,
        "prerelease": prerelease,
    }

    with requests_mock.Mocker(session=default_gitea_client.session) as m:
        # mock the response
        m.register_uri(
            "POST",
            gitea_api_matcher,
            json={"id": mock_release_id},
            status_code=status_code,
        )

        # Execute method under test expecting an exeception to be raised
        with pytest.raises(HTTPError):
            default_gitea_client.create_release(tag, RELEASE_NOTES, prerelease)

        # Evaluate (expected -> actual)
        assert m.called
        assert expected_num_requests == len(m.request_history)
        assert expected_http_method == m.last_request.method
        assert expected_request_url == m.last_request.url
        assert expected_request_body == m.last_request.json()


@pytest.mark.parametrize("token", (None, "super-token"))
def test_should_create_release_using_token_or_netrc(
    default_gitea_client: Gitea,
    token: str | None,
    default_netrc_username: str,
    default_netrc_password: str,
    netrc_file: NetrcFileFn,
    clean_os_environment: dict[str, str],
):
    # Setup
    default_gitea_client.token = token
    default_gitea_client.session.auth = None if not token else TokenAuth(token)
    tag = "v1.0.0"
    expected_release_id = 1
    expected_num_requests = 1
    expected_http_method = "POST"
    expected_request_url = "{api_url}/repos/{owner}/{repo_name}/releases".format(
        api_url=default_gitea_client.api_url,
        owner=default_gitea_client.owner,
        repo_name=default_gitea_client.repo_name,
    )
    expected_request_body = {
        "tag_name": tag,
        "name": tag,
        "body": RELEASE_NOTES,
        "draft": False,
        "prerelease": False,
    }

    expected_request_headers = set(
        (
            {"Authorization": f"token {token}"}
            if token
            else {
                "Authorization": _basic_auth_str(
                    default_netrc_username, default_netrc_password
                )
            }
        ).items()
    )

    # create netrc file
    # NOTE: write netrc file with DEFAULT_DOMAIN not DEFAULT_API_DOMAIN as can't
    #       handle /api/v1 in file
    netrc = netrc_file(machine=default_gitea_client.DEFAULT_DOMAIN)

    mocked_os_environ = {**clean_os_environment, "NETRC": netrc.name}

    # Monkeypatch to create the Mocked environment
    with requests_mock.Mocker(
        session=default_gitea_client.session
    ) as m, mock.patch.dict(os.environ, mocked_os_environ, clear=True):
        # mock the response
        m.register_uri(
            "POST", gitea_api_matcher, json={"id": expected_release_id}, status_code=201
        )

        # Execute method under test
        ret_val = default_gitea_client.create_release(tag, RELEASE_NOTES)

        # Evaluate (expected -> actual)
        assert expected_release_id == ret_val
        assert m.called
        assert expected_num_requests == len(m.request_history)
        assert expected_http_method == m.last_request.method
        assert expected_request_url == m.last_request.url
        assert expected_request_body == m.last_request.json()

        # calculate the match between expected and actual headers
        # We are not looking for an exact match, just that the headers we must have exist
        shared_headers = expected_request_headers.intersection(
            set(m.last_request.headers.items())
        )
        assert expected_request_headers == shared_headers, str.join(
            os.linesep,
            [
                "Actual headers are missing some of the expected headers",
                f"Matching: {shared_headers}",
                f"Missing: {expected_request_headers - shared_headers}",
                f"Extra: {set(m.last_request.headers.items()) - expected_request_headers}",
            ],
        )


def test_request_has_no_auth_header_if_no_token_or_netrc():
    tag = "v1.0.0"
    expected_release_id = 1
    expected_num_requests = 1
    expected_http_method = "POST"

    with mock.patch.dict(os.environ, {}, clear=True):
        client = Gitea(remote_url=f"git@{Gitea.DEFAULT_DOMAIN}:something/somewhere.git")

        expected_request_url = "{api_url}/repos/{owner}/{repo_name}/releases".format(
            api_url=client.api_url,
            owner=client.owner,
            repo_name=client.repo_name,
        )

        with requests_mock.Mocker(session=client.session) as m:
            # mock the response
            m.register_uri("POST", gitea_api_matcher, json={"id": 1}, status_code=201)

            # Execute method under test
            ret_val = client.create_release(tag, RELEASE_NOTES)

            # Evaluate (expected -> actual)
            assert expected_release_id == ret_val
            assert m.called
            assert expected_num_requests == len(m.request_history)
            assert expected_http_method == m.last_request.method
            assert expected_request_url == m.last_request.url
            assert "Authorization" not in m.last_request.headers


@pytest.mark.parametrize(
    "resp_payload, status_code, expected_result",
    [
        ({"id": 420}, 200, 420),
        ({}, 404, None),
    ],
)
def test_get_release_id_by_tag(
    default_gitea_client: Gitea,
    resp_payload: dict[str, int],
    status_code: int,
    expected_result: int | None,
):
    # Setup
    tag = "v1.0.0"
    expected_num_requests = 1
    expected_http_method = "GET"
    expected_request_url = (
        "{api_url}/repos/{owner}/{repo_name}/releases/tags/{tag}".format(
            api_url=default_gitea_client.api_url,
            owner=default_gitea_client.owner,
            repo_name=default_gitea_client.repo_name,
            tag=tag,
        )
    )

    with requests_mock.Mocker(session=default_gitea_client.session) as m:
        # mock the response
        m.register_uri(
            "GET", gitea_api_matcher, json=resp_payload, status_code=status_code
        )

        # Execute method under test
        rtn_val = default_gitea_client.get_release_id_by_tag(tag)

        # Evaluate (expected -> actual)
        assert expected_result == rtn_val
        assert m.called
        assert expected_num_requests == len(m.request_history)
        assert expected_http_method == m.last_request.method
        assert expected_request_url == m.last_request.url


@pytest.mark.parametrize("status_code", [201])
@pytest.mark.parametrize("mock_release_id", range(3))
def test_edit_release_notes_succeeds(
    default_gitea_client: Gitea,
    status_code: int,
    mock_release_id: int,
):
    # Setup
    expected_num_requests = 1
    expected_http_method = "PATCH"
    expected_request_url = (
        "{api_url}/repos/{owner}/{repo_name}/releases/{release_id}".format(
            api_url=default_gitea_client.api_url,
            owner=default_gitea_client.owner,
            repo_name=default_gitea_client.repo_name,
            release_id=mock_release_id,
        )
    )
    expected_request_body = {"body": RELEASE_NOTES}

    with requests_mock.Mocker(session=default_gitea_client.session) as m:
        # mock the response
        m.register_uri(
            "PATCH",
            gitea_api_matcher,
            json={"id": mock_release_id},
            status_code=status_code,
        )

        # Execute method under test
        rtn_val = default_gitea_client.edit_release_notes(
            mock_release_id, RELEASE_NOTES
        )

        # Evaluate (expected -> actual)
        assert mock_release_id == rtn_val
        assert m.called
        assert expected_num_requests == len(m.request_history)
        assert expected_http_method == m.last_request.method
        assert expected_request_url == m.last_request.url
        assert expected_request_body == m.last_request.json()


@pytest.mark.parametrize("status_code", (400, 404, 429, 500, 503))
@pytest.mark.parametrize("mock_release_id", range(3))
def test_edit_release_notes_fails(
    default_gitea_client: Gitea,
    status_code: int,
    mock_release_id: int,
):
    # Setup
    expected_num_requests = 1
    expected_http_method = "PATCH"
    expected_request_url = (
        "{api_url}/repos/{owner}/{repo_name}/releases/{release_id}".format(
            api_url=default_gitea_client.api_url,
            owner=default_gitea_client.owner,
            repo_name=default_gitea_client.repo_name,
            release_id=mock_release_id,
        )
    )
    expected_request_body = {"body": RELEASE_NOTES}

    with requests_mock.Mocker(session=default_gitea_client.session) as m:
        # mock the response
        m.register_uri(
            "PATCH",
            gitea_api_matcher,
            json={"id": mock_release_id},
            status_code=status_code,
        )

        # Execute method under test expecting an exception to be raised
        with pytest.raises(HTTPError):
            default_gitea_client.edit_release_notes(mock_release_id, RELEASE_NOTES)

        assert m.called
        assert expected_num_requests == len(m.request_history)
        assert expected_http_method == m.last_request.method
        assert expected_request_url == m.last_request.url
        assert expected_request_body == m.last_request.json()


# Note - mocking as the logic for the create/update of a release
# is covered by testing above, no point re-testing.


@pytest.mark.parametrize("mock_release_id", range(3))
@pytest.mark.parametrize("prerelease", (True, False))
def test_create_or_update_release_when_create_succeeds(
    default_gitea_client: Gitea,
    mock_release_id: int,
    prerelease: bool,
):
    tag = "v1.0.0"
    with mock.patch.object(
        default_gitea_client,
        default_gitea_client.create_release.__name__,
        return_value=mock_release_id,
    ) as mock_create_release, mock.patch.object(
        default_gitea_client,
        default_gitea_client.get_release_id_by_tag.__name__,
        return_value=mock_release_id,
    ) as mock_get_release_id_by_tag, mock.patch.object(
        default_gitea_client,
        default_gitea_client.edit_release_notes.__name__,
        return_value=mock_release_id,
    ) as mock_edit_release_notes:
        # Execute in mock environment
        result = default_gitea_client.create_or_update_release(
            tag, RELEASE_NOTES, prerelease
        )

        # Evaluate (expected -> actual)
        assert mock_release_id == result
        mock_create_release.assert_called_once_with(tag, RELEASE_NOTES, prerelease)
        mock_get_release_id_by_tag.assert_not_called()
        mock_edit_release_notes.assert_not_called()


@pytest.mark.parametrize("mock_release_id", range(3))
@pytest.mark.parametrize("prerelease", (True, False))
def test_create_or_update_release_when_create_fails_and_update_succeeds(
    default_gitea_client: Gitea,
    mock_release_id: int,
    prerelease: bool,
):
    tag = "v1.0.0"
    not_found = HTTPError("404 Not Found")
    not_found.response = Response()
    not_found.response.status_code = 404

    with mock.patch.object(
        default_gitea_client,
        default_gitea_client.create_release.__name__,
        side_effect=not_found,
    ) as mock_create_release, mock.patch.object(
        default_gitea_client,
        default_gitea_client.get_release_id_by_tag.__name__,
        return_value=mock_release_id,
    ) as mock_get_release_id_by_tag, mock.patch.object(
        default_gitea_client,
        default_gitea_client.edit_release_notes.__name__,
        return_value=mock_release_id,
    ) as mock_edit_release_notes:
        # Execute in mock environment
        result = default_gitea_client.create_or_update_release(
            tag, RELEASE_NOTES, prerelease
        )

        # Evaluate (expected -> actual)
        assert mock_release_id == result
        mock_create_release.assert_called_once()
        mock_get_release_id_by_tag.assert_called_once_with(tag)
        mock_edit_release_notes.assert_called_once_with(mock_release_id, RELEASE_NOTES)


@pytest.mark.parametrize("prerelease", (True, False))
def test_create_or_update_release_when_create_fails_and_no_release_for_tag(
    default_gitea_client: Gitea,
    prerelease: bool,
):
    tag = "v1.0.0"
    not_found = HTTPError("404 Not Found")
    not_found.response = Response()
    not_found.response.status_code = 404

    with mock.patch.object(
        default_gitea_client,
        default_gitea_client.create_release.__name__,
        side_effect=not_found,
    ) as mock_create_release, mock.patch.object(
        default_gitea_client,
        default_gitea_client.get_release_id_by_tag.__name__,
        return_value=None,
    ) as mock_get_release_id_by_tag, mock.patch.object(
        default_gitea_client,
        default_gitea_client.edit_release_notes.__name__,
        return_value=None,
    ) as mock_edit_release_notes:
        # Execute in mock environment expecting an exception to be raised
        with pytest.raises(ValueError):
            default_gitea_client.create_or_update_release(
                tag, RELEASE_NOTES, prerelease
            )

        mock_create_release.assert_called_once()
        mock_get_release_id_by_tag.assert_called_once_with(tag)
        mock_edit_release_notes.assert_not_called()


@pytest.mark.parametrize("status_code", (200, 201))
@pytest.mark.parametrize("mock_release_id", range(3))
@pytest.mark.usefixtures(init_example_project.__name__)
def test_upload_release_asset_succeeds(
    default_gitea_client: Gitea,
    example_changelog_md: Path,
    status_code: int,
    mock_release_id: int,
):
    # Setup
    urlparams = {"name": example_changelog_md.name}
    expected_num_requests = 1
    expected_http_method = "POST"
    expected_request_url = "{url}?{params}".format(
        url=default_gitea_client.asset_upload_url(mock_release_id),
        params=urlencode(urlparams),
    )
    expected_changelog = example_changelog_md.read_bytes()

    with requests_mock.Mocker(session=default_gitea_client.session) as m:
        m.register_uri(
            "POST", gitea_api_matcher, json={"status": "ok"}, status_code=status_code
        )
        result = default_gitea_client.upload_release_asset(
            release_id=mock_release_id,
            file=example_changelog_md.resolve(),
            label="doesn't matter could be None",
        )

        # Evaluate (expected -> actual)
        assert result is True
        assert m.called
        assert expected_num_requests == len(m.request_history)
        assert expected_http_method == m.last_request.method
        assert expected_request_url == m.last_request.url
        assert expected_changelog in m.last_request.body


@pytest.mark.parametrize("status_code", (400, 500, 503))
@pytest.mark.parametrize("mock_release_id", range(3))
@pytest.mark.usefixtures(init_example_project.__name__)
def test_upload_release_asset_fails(
    default_gitea_client: Gitea,
    example_changelog_md: Path,
    status_code: int,
    mock_release_id: int,
):
    with requests_mock.Mocker(session=default_gitea_client.session) as m:
        # mock the response
        m.register_uri(
            "POST", gitea_api_matcher, json={"status": "error"}, status_code=status_code
        )

        # Execute method under test expecting an exception to be raised
        with pytest.raises(HTTPError):
            default_gitea_client.upload_release_asset(
                release_id=mock_release_id,
                file=example_changelog_md.resolve(),
                label="doesn't matter could be None",
            )


# Note - mocking as the logic for uploading an asset
# is covered by testing above, no point re-testing.
def test_upload_dists_when_release_id_not_found(default_gitea_client: Gitea):
    tag = "v1.0.0"
    path = "doesn't matter"
    expected_num_uploads = 0

    # Set up mock environment
    with mock.patch.object(
        default_gitea_client,
        default_gitea_client.get_release_id_by_tag.__name__,
        return_value=None,
    ) as mock_get_release_id_by_tag, mock.patch.object(
        default_gitea_client, default_gitea_client.upload_release_asset.__name__
    ) as mock_upload_release_asset:
        # Execute method under test
        result = default_gitea_client.upload_dists(tag, path)

        # Evaluate
        assert expected_num_uploads == result
        mock_get_release_id_by_tag.assert_called_once_with(tag=tag)
        mock_upload_release_asset.assert_not_called()


@pytest.mark.parametrize(
    "files, glob_pattern, upload_statuses, expected_num_uploads",
    [
        (["foo.zip", "bar.whl"], "*.zip", [True], 1),
        (["foo.whl", "foo.egg", "foo.tar.gz"], "foo.*", [True, True, True], 3),
        # What if not built?
        ([], "*", [], 0),
        # What if wrong directory/other stuff in output dir/subfolder?
        (["specialconfig.yaml", "something.whl", "desc.md"], "*.yaml", [True], 1),
        (["specialconfig.yaml", "something.whl", "desc.md"], "*.md", [True], 1),
    ],
)
def test_upload_dists_when_release_id_found(
    default_gitea_client: Gitea,
    files: list[str],
    glob_pattern: str,
    upload_statuses: list[bool],
    expected_num_uploads: int,
):
    release_id = 420
    tag = "doesn't matter"
    matching_files = fnmatch.filter(files, glob_pattern)
    expected_files_uploaded = [mock.call(release_id, fn) for fn in matching_files]

    # Skip check as the files don't exist in filesystem
    mocked_isfile = mock.patch.object(os.path, "isfile", return_value=True)
    mocked_globber = mock.patch.object(glob, "glob", return_value=matching_files)

    # Set up mock environment
    with mocked_globber, mocked_isfile, mock.patch.object(
        default_gitea_client,
        default_gitea_client.get_release_id_by_tag.__name__,
        return_value=release_id,
    ) as mock_get_release_id_by_tag, mock.patch.object(
        default_gitea_client,
        default_gitea_client.upload_release_asset.__name__,
        side_effect=upload_statuses,
    ) as mock_upload_release_asset:
        # Execute method under test
        num_uploads = default_gitea_client.upload_dists(tag, glob_pattern)

        # Evaluate (expected -> actual)
        assert expected_num_uploads == num_uploads
        mock_get_release_id_by_tag.assert_called_once_with(tag=tag)
        assert expected_files_uploaded == mock_upload_release_asset.call_args_list



from tests.fixtures.repos.github_flow.repo_w_default_release import *
from tests.fixtures.repos.github_flow.repo_w_default_release_w_branch_update_merge import *
from tests.fixtures.repos.github_flow.repo_w_release_channels import *



from __future__ import annotations

import fnmatch
import glob
import os
import re
from typing import TYPE_CHECKING
from unittest import mock
from urllib.parse import urlencode

import pytest
import requests_mock
from requests import HTTPError, Response, Session
from requests.auth import _basic_auth_str

from semantic_release.hvcs.github import Github
from semantic_release.hvcs.token_auth import TokenAuth

from tests.const import (
    EXAMPLE_HVCS_DOMAIN,
    EXAMPLE_REPO_NAME,
    EXAMPLE_REPO_OWNER,
    RELEASE_NOTES,
)
from tests.fixtures.example_project import init_example_project

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Generator

    from tests.conftest import NetrcFileFn


@pytest.fixture
def default_gh_client() -> Generator[Github, None, None]:
    remote_url = (
        f"git@{Github.DEFAULT_DOMAIN}:{EXAMPLE_REPO_OWNER}/{EXAMPLE_REPO_NAME}.git"
    )
    with mock.patch.dict(os.environ, {}, clear=True):
        yield Github(remote_url=remote_url)


@pytest.mark.parametrize(
    str.join(
        ", ",
        [
            "patched_os_environ",
            "hvcs_domain",
            "hvcs_api_domain",
            "expected_hvcs_domain",
            "expected_hvcs_api_url",
            "insecure",
        ],
    ),
    [
        (
            # Default values (GitHub Enterprise Cloud)
            {},
            None,
            None,
            "https://github.com",
            "https://api.github.com",
            False,
        ),
        (
            # Explicitly set default values (GitHub Enterprise Cloud)
            {},
            Github.DEFAULT_DOMAIN,
            Github.DEFAULT_API_DOMAIN,
            "https://github.com",
            "https://api.github.com",
            False,
        ),
        (
            # Pull both locations from environment (GitHub Actions on Cloud)
            {
                "GITHUB_SERVER_URL": f"https://{Github.DEFAULT_DOMAIN}",
                "GITHUB_API_URL": f"https://{Github.DEFAULT_API_DOMAIN}",
            },
            None,
            None,
            "https://github.com",
            "https://api.github.com",
            False,
        ),
        (
            # Explicitly set custom values with full api path
            {},
            EXAMPLE_HVCS_DOMAIN,
            f"{EXAMPLE_HVCS_DOMAIN}{Github.DEFAULT_API_PATH_ONPREM}",
            f"https://{EXAMPLE_HVCS_DOMAIN}",
            f"https://{EXAMPLE_HVCS_DOMAIN}{Github.DEFAULT_API_PATH_ONPREM}",
            False,
        ),
        (
            # Explicitly defined api as subdomain
            # POSSIBLY WRONG ASSUMPTION of Api path for GitHub Enterprise Server (On Prem)
            {},
            f"https://{EXAMPLE_HVCS_DOMAIN}",
            f"https://api.{EXAMPLE_HVCS_DOMAIN}",
            f"https://{EXAMPLE_HVCS_DOMAIN}",
            f"https://api.{EXAMPLE_HVCS_DOMAIN}{Github.DEFAULT_API_PATH_ONPREM}",
            False,
        ),
        (
            # Custom domain with path prefix
            {},
            "special.custom.server/vcs",
            None,
            "https://special.custom.server/vcs",
            "https://special.custom.server/vcs/api/v3",
            False,
        ),
        (
            # Gather domain from environment & imply api domain from server domain
            {"GITHUB_SERVER_URL": "https://special.custom.server/"},
            None,
            None,
            "https://special.custom.server",
            "https://special.custom.server/api/v3",
            False,
        ),
        (
            # Pull both locations from environment (On-prem Actions Env)
            {
                "GITHUB_SERVER_URL": "https://special.custom.server/",
                "GITHUB_API_URL": "https://special.custom.server/api/v3",
            },
            None,
            None,
            "https://special.custom.server",
            "https://special.custom.server/api/v3",
            False,
        ),
        (
            # Ignore environment & use provided parameter value (ie from user config)
            # then infer api domain from the parameter value based on default GitHub configurations
            {"GITHUB_SERVER_URL": "https://special.custom.server/vcs/"},
            f"https://{EXAMPLE_HVCS_DOMAIN}",
            None,
            f"https://{EXAMPLE_HVCS_DOMAIN}",
            f"https://{EXAMPLE_HVCS_DOMAIN}/api/v3",
            False,
        ),
        (
            # Ignore environment & use provided parameter value (ie from user config)
            {"GITHUB_API_URL": "https://api.special.custom.server/"},
            f"https://{EXAMPLE_HVCS_DOMAIN}",
            f"https://{EXAMPLE_HVCS_DOMAIN}/api/v3",
            f"https://{EXAMPLE_HVCS_DOMAIN}",
            f"https://{EXAMPLE_HVCS_DOMAIN}/api/v3",
            False,
        ),
        (
            # Allow insecure http connections explicitly
            {},
            f"http://{EXAMPLE_HVCS_DOMAIN}",
            f"http://{EXAMPLE_HVCS_DOMAIN}/api/v3",
            f"http://{EXAMPLE_HVCS_DOMAIN}",
            f"http://{EXAMPLE_HVCS_DOMAIN}/api/v3",
            True,
        ),
        (
            # Allow insecure http connections explicitly & imply insecure api domain
            {},
            f"http://{EXAMPLE_HVCS_DOMAIN}",
            None,
            f"http://{EXAMPLE_HVCS_DOMAIN}",
            f"http://{EXAMPLE_HVCS_DOMAIN}/api/v3",
            True,
        ),
        (
            # Infer insecure connection from user configuration
            {},
            EXAMPLE_HVCS_DOMAIN,
            EXAMPLE_HVCS_DOMAIN,
            f"http://{EXAMPLE_HVCS_DOMAIN}",
            f"http://{EXAMPLE_HVCS_DOMAIN}/api/v3",
            True,
        ),
        (
            # Infer insecure connection from user configuration & imply insecure api domain
            {},
            EXAMPLE_HVCS_DOMAIN,
            None,
            f"http://{EXAMPLE_HVCS_DOMAIN}",
            f"http://{EXAMPLE_HVCS_DOMAIN}/api/v3",
            True,
        ),
    ],
)
@pytest.mark.parametrize(
    "remote_url",
    [
        f"git@{Github.DEFAULT_DOMAIN}:{EXAMPLE_REPO_OWNER}/{EXAMPLE_REPO_NAME}.git",
        f"https://{Github.DEFAULT_DOMAIN}/{EXAMPLE_REPO_OWNER}/{EXAMPLE_REPO_NAME}.git",
    ],
)
@pytest.mark.parametrize("token", ("abc123", None))
def test_github_client_init(
    patched_os_environ: dict[str, str],
    hvcs_domain: str | None,
    hvcs_api_domain: str | None,
    expected_hvcs_domain: str,
    expected_hvcs_api_url: str,
    remote_url: str,
    token: str | None,
    insecure: bool,
):
    with mock.patch.dict(os.environ, patched_os_environ, clear=True):
        client = Github(
            remote_url=remote_url,
            hvcs_domain=hvcs_domain,
            hvcs_api_domain=hvcs_api_domain,
            token=token,
            allow_insecure=insecure,
        )

        # Evaluate (expected -> actual)
        assert expected_hvcs_domain == str(client.hvcs_domain)
        assert expected_hvcs_api_url == str(client.api_url)
        assert token == client.token
        assert remote_url == client._remote_url
        assert hasattr(client, "session")
        assert isinstance(getattr(client, "session", None), Session)


@pytest.mark.parametrize(
    "hvcs_domain, hvcs_api_domain, insecure",
    [
        # Bad base domain schemes
        (f"ftp://{EXAMPLE_HVCS_DOMAIN}", None, False),
        (f"ftp://{EXAMPLE_HVCS_DOMAIN}", None, True),
        # Unallowed insecure connections when base domain is insecure
        (f"http://{EXAMPLE_HVCS_DOMAIN}", None, False),
        # Bad API domain schemes
        (None, f"ftp://api.{EXAMPLE_HVCS_DOMAIN}", False),
        (None, f"ftp://api.{EXAMPLE_HVCS_DOMAIN}", True),
        # Unallowed insecure connections when api domain is insecure
        (None, f"http://{EXAMPLE_HVCS_DOMAIN}", False),
    ],
)
def test_github_client_init_with_invalid_scheme(
    hvcs_domain: str | None,
    hvcs_api_domain: str | None,
    insecure: bool,
):
    with pytest.raises(ValueError), mock.patch.dict(os.environ, {}, clear=True):
        Github(
            remote_url=f"https://{EXAMPLE_HVCS_DOMAIN}/{EXAMPLE_REPO_OWNER}/{EXAMPLE_REPO_NAME}.git",
            hvcs_domain=hvcs_domain,
            hvcs_api_domain=hvcs_api_domain,
            allow_insecure=insecure,
        )


@pytest.mark.parametrize(
    "patched_os_environ, expected_owner, expected_name",
    [
        ({}, None, None),
        ({"GITHUB_REPOSITORY": "path/to/repo/foo"}, "path/to/repo", "foo"),
    ],
)
def test_github_get_repository_owner_and_name(
    default_gh_client: Github,
    patched_os_environ: dict[str, str],
    expected_owner: str,
    expected_name: str,
):
    # expected results should be a tuple[namespace, repo_name]
    # when None, the default values are used which matches default_gh_client's setup
    expected_result = (
        expected_owner or EXAMPLE_REPO_OWNER,
        expected_name or EXAMPLE_REPO_NAME,
    )

    with mock.patch.dict(os.environ, patched_os_environ, clear=True):
        # Execute in mocked environment
        result = default_gh_client._get_repository_owner_and_name()

        # Evaluate (expected -> actual)
        assert expected_result == result


def test_compare_url(default_gh_client: Github):
    # Setup
    start_rev = "revA"
    end_rev = "revB"
    expected_url = "{server}/{owner}/{repo}/compare/{from_rev}...{to_rev}".format(
        server=default_gh_client.hvcs_domain,
        owner=default_gh_client.owner,
        repo=default_gh_client.repo_name,
        from_rev=start_rev,
        to_rev=end_rev,
    )

    # Execute method under test
    actual_url = default_gh_client.compare_url(from_rev=start_rev, to_rev=end_rev)

    # Evaluate (expected -> actual)
    assert expected_url == actual_url


@pytest.mark.parametrize(
    "patched_os_environ, use_token, token, remote_url, expected_auth_url",
    [
        (
            {"GITHUB_ACTOR": "foo"},
            False,
            "",
            f"git@{Github.DEFAULT_DOMAIN}:custom/example.git",
            f"git@{Github.DEFAULT_DOMAIN}:custom/example.git",
        ),
        (
            {"GITHUB_ACTOR": "foo"},
            True,
            "",
            f"git@{Github.DEFAULT_DOMAIN}:custom/example.git",
            f"git@{Github.DEFAULT_DOMAIN}:custom/example.git",
        ),
        (
            {},
            False,
            "aabbcc",
            f"git@{Github.DEFAULT_DOMAIN}:custom/example.git",
            f"git@{Github.DEFAULT_DOMAIN}:custom/example.git",
        ),
        (
            {},
            True,
            "aabbcc",
            f"git@{Github.DEFAULT_DOMAIN}:custom/example.git",
            f"https://aabbcc@{Github.DEFAULT_DOMAIN}/custom/example.git",
        ),
        (
            {"GITHUB_ACTOR": "foo"},
            False,
            "aabbcc",
            f"git@{Github.DEFAULT_DOMAIN}:custom/example.git",
            f"git@{Github.DEFAULT_DOMAIN}:custom/example.git",
        ),
        (
            {"GITHUB_ACTOR": "foo"},
            True,
            "aabbcc",
            f"git@{Github.DEFAULT_DOMAIN}:custom/example.git",
            f"https://foo:aabbcc@{Github.DEFAULT_DOMAIN}/custom/example.git",
        ),
    ],
)
def test_remote_url(
    default_gh_client: Github,
    patched_os_environ: dict[str, str],
    use_token: bool,
    token: str,
    remote_url: str,
    expected_auth_url: str,
):
    with mock.patch.dict(os.environ, patched_os_environ, clear=True):
        default_gh_client._remote_url = remote_url
        default_gh_client.token = token

        # Execute method under test & Evaluate (expected -> actual)
        assert expected_auth_url == default_gh_client.remote_url(use_token=use_token)


def test_commit_hash_url(default_gh_client: Github):
    sha = "hashashash"
    expected_url = "{server}/{owner}/{repo}/commit/{sha}".format(
        server=default_gh_client.hvcs_domain.url,
        owner=default_gh_client.owner,
        repo=default_gh_client.repo_name,
        sha=sha,
    )
    assert expected_url == default_gh_client.commit_hash_url(sha)


def test_commit_hash_url_w_custom_server():
    """
    Test the commit hash URL generation for a self-hosted Bitbucket server with prefix.

    ref: https://github.com/python-semantic-release/python-semantic-release/issues/1204
    """
    sha = "244f7e11bcb1e1ce097db61594056bc2a32189a0"
    expected_url = "{server}/{owner}/{repo}/commit/{sha}".format(
        server=f"https://{EXAMPLE_HVCS_DOMAIN}/projects/demo-foo",
        owner="foo",
        repo=EXAMPLE_REPO_NAME,
        sha=sha,
    )

    with mock.patch.dict(os.environ, {}, clear=True):
        actual_url = Github(
            remote_url=f"https://{EXAMPLE_HVCS_DOMAIN}/projects/demo-foo/foo/{EXAMPLE_REPO_NAME}.git",
            hvcs_domain=f"https://{EXAMPLE_HVCS_DOMAIN}/projects/demo-foo",
        ).commit_hash_url(sha)

    assert expected_url == actual_url


@pytest.mark.parametrize("issue_number", (666, "666", "#666"))
def test_issue_url(default_gh_client: Github, issue_number: str | int):
    expected_url = "{server}/{owner}/{repo}/issues/{issue_num}".format(
        server=default_gh_client.hvcs_domain.url,
        owner=default_gh_client.owner,
        repo=default_gh_client.repo_name,
        issue_num=str(issue_number).lstrip("#"),
    )
    assert expected_url == default_gh_client.issue_url(issue_num=issue_number)


@pytest.mark.parametrize("pr_number", (666, "666", "#666"))
def test_pull_request_url(default_gh_client: Github, pr_number: int | str):
    expected_url = "{server}/{owner}/{repo}/pull/{pr_number}".format(
        server=default_gh_client.hvcs_domain,
        owner=default_gh_client.owner,
        repo=default_gh_client.repo_name,
        pr_number=str(pr_number).lstrip("#"),
    )
    actual_url = default_gh_client.pull_request_url(pr_number=pr_number)
    assert expected_url == actual_url


############
# Tests which need http response mocking
############


github_upload_url = f"https://uploads.{Github.DEFAULT_DOMAIN}"
github_matcher = re.compile(rf"^https://{Github.DEFAULT_DOMAIN}")
github_api_matcher = re.compile(rf"^https://{Github.DEFAULT_API_DOMAIN}")
github_upload_matcher = re.compile(rf"^{github_upload_url}")


@pytest.mark.parametrize("status_code", (200, 201))
@pytest.mark.parametrize("mock_release_id", range(3))
@pytest.mark.parametrize("prerelease", (True, False))
def test_create_release_succeeds(
    default_gh_client: Github,
    mock_release_id: int,
    prerelease: bool,
    status_code: int,
):
    tag = "v1.0.0"
    expected_num_requests = 1
    expected_http_method = "POST"
    expected_request_url = "{api_url}/repos/{owner}/{repo_name}/releases".format(
        api_url=default_gh_client.api_url,
        owner=default_gh_client.owner,
        repo_name=default_gh_client.repo_name,
    )
    expected_request_body = {
        "tag_name": tag,
        "name": tag,
        "body": RELEASE_NOTES,
        "draft": False,
        "prerelease": prerelease,
    }

    with requests_mock.Mocker(session=default_gh_client.session) as m:
        # mock the response
        m.register_uri(
            "POST",
            github_api_matcher,
            json={"id": mock_release_id},
            status_code=status_code,
        )

        # Execute method under test
        actual_rtn_val = default_gh_client.create_release(
            tag, RELEASE_NOTES, prerelease
        )

        # Evaluate (expected -> actual)
        assert mock_release_id == actual_rtn_val
        assert m.called
        assert expected_num_requests == len(m.request_history)
        assert expected_http_method == m.last_request.method
        assert expected_request_url == m.last_request.url
        assert expected_request_body == m.last_request.json()


@pytest.mark.parametrize("status_code", (400, 404, 429, 500, 503))
@pytest.mark.parametrize("mock_release_id", range(3))
@pytest.mark.parametrize("prerelease", (True, False))
def test_create_release_fails(
    default_gh_client: Github,
    mock_release_id: int,
    prerelease: bool,
    status_code: int,
):
    tag = "v1.0.0"
    expected_num_requests = 1
    expected_http_method = "POST"
    expected_request_url = "{api_url}/repos/{owner}/{repo_name}/releases".format(
        api_url=default_gh_client.api_url,
        owner=default_gh_client.owner,
        repo_name=default_gh_client.repo_name,
    )
    expected_request_body = {
        "tag_name": tag,
        "name": tag,
        "body": RELEASE_NOTES,
        "draft": False,
        "prerelease": prerelease,
    }

    with requests_mock.Mocker(session=default_gh_client.session) as m:
        # mock the response
        m.register_uri(
            "POST",
            github_api_matcher,
            json={"id": mock_release_id},
            status_code=status_code,
        )

        # Execute method under test expecting an exeception to be raised
        with pytest.raises(HTTPError):
            default_gh_client.create_release(tag, RELEASE_NOTES, prerelease)

        # Evaluate (expected -> actual)
        assert m.called
        assert expected_num_requests == len(m.request_history)
        assert expected_http_method == m.last_request.method
        assert expected_request_url == m.last_request.url
        assert expected_request_body == m.last_request.json()


@pytest.mark.parametrize("token", (None, "super-token"))
def test_should_create_release_using_token_or_netrc(
    default_gh_client: Github,
    token: str | None,
    default_netrc_username: str,
    default_netrc_password: str,
    netrc_file: NetrcFileFn,
    clean_os_environment: dict[str, str],
):
    # Setup
    default_gh_client.token = token
    default_gh_client.session.auth = None if not token else TokenAuth(token)
    tag = "v1.0.0"
    expected_release_id = 1
    expected_num_requests = 1
    expected_http_method = "POST"
    expected_request_url = "{api_url}/repos/{owner}/{repo_name}/releases".format(
        api_url=default_gh_client.api_url,
        owner=default_gh_client.owner,
        repo_name=default_gh_client.repo_name,
    )
    expected_request_body = {
        "tag_name": tag,
        "name": tag,
        "body": RELEASE_NOTES,
        "draft": False,
        "prerelease": False,
    }

    expected_request_headers = set(
        (
            {"Authorization": f"token {token}"}
            if token
            else {
                "Authorization": _basic_auth_str(
                    default_netrc_username, default_netrc_password
                )
            }
        ).items()
    )

    # create netrc file
    netrc = netrc_file(machine=default_gh_client.DEFAULT_API_DOMAIN)

    mocked_os_environ = {**clean_os_environment, "NETRC": netrc.name}

    # Monkeypatch to create the Mocked environment
    with requests_mock.Mocker(session=default_gh_client.session) as m, mock.patch.dict(
        os.environ, mocked_os_environ, clear=True
    ):
        # mock the response
        m.register_uri(
            "POST",
            github_api_matcher,
            json={"id": expected_release_id},
            status_code=201,
        )

        # Execute method under test
        ret_val = default_gh_client.create_release(tag, RELEASE_NOTES)

        # Evaluate (expected -> actual)
        assert expected_release_id == ret_val
        assert m.called
        assert expected_num_requests == len(m.request_history)
        assert expected_http_method == m.last_request.method
        assert expected_request_url == m.last_request.url
        assert expected_request_body == m.last_request.json()

        # calculate the match between expected and actual headers
        # We are not looking for an exact match, just that the headers we must have exist
        shared_headers = expected_request_headers.intersection(
            set(m.last_request.headers.items())
        )
        assert expected_request_headers == shared_headers, str.join(
            os.linesep,
            [
                "Actual headers are missing some of the expected headers",
                f"Matching: {shared_headers}",
                f"Missing: {expected_request_headers - shared_headers}",
                f"Extra: {set(m.last_request.headers.items()) - expected_request_headers}",
            ],
        )


def test_request_has_no_auth_header_if_no_token_or_netrc():
    tag = "v1.0.0"
    expected_release_id = 1
    expected_num_requests = 1
    expected_http_method = "POST"

    with mock.patch.dict(os.environ, {}, clear=True):
        client = Github(
            remote_url=f"git@{Github.DEFAULT_DOMAIN}:something/somewhere.git"
        )

        expected_request_url = "{api_url}/repos/{owner}/{repo_name}/releases".format(
            api_url=client.api_url,
            owner=client.owner,
            repo_name=client.repo_name,
        )

        with requests_mock.Mocker(session=client.session) as m:
            # mock the response
            m.register_uri("POST", github_api_matcher, json={"id": 1}, status_code=201)

            # Execute method under test
            rtn_val = client.create_release(tag, RELEASE_NOTES)

            # Evaluate (expected -> actual)
            assert expected_release_id == rtn_val
            assert m.called
            assert expected_num_requests == len(m.request_history)
            assert expected_http_method == m.last_request.method
            assert expected_request_url == m.last_request.url
            assert "Authorization" not in m.last_request.headers


@pytest.mark.parametrize("status_code", [201])
@pytest.mark.parametrize("mock_release_id", range(3))
def test_edit_release_notes_succeeds(
    default_gh_client: Github,
    status_code: int,
    mock_release_id: int,
):
    # Setup
    expected_num_requests = 1
    expected_http_method = "POST"
    expected_request_url = (
        "{api_url}/repos/{owner}/{repo_name}/releases/{release_id}".format(
            api_url=default_gh_client.api_url,
            owner=default_gh_client.owner,
            repo_name=default_gh_client.repo_name,
            release_id=mock_release_id,
        )
    )
    expected_request_body = {"body": RELEASE_NOTES}

    with requests_mock.Mocker(session=default_gh_client.session) as m:
        # mock the response
        m.register_uri(
            "POST",
            github_api_matcher,
            json={"id": mock_release_id},
            status_code=status_code,
        )

        # Execute method under test
        rtn_val = default_gh_client.edit_release_notes(mock_release_id, RELEASE_NOTES)

        # Evaluate (expected -> actual)
        assert mock_release_id == rtn_val
        assert m.called
        assert expected_num_requests == len(m.request_history)
        assert expected_http_method == m.last_request.method
        assert expected_request_url == m.last_request.url
        assert expected_request_body == m.last_request.json()


@pytest.mark.parametrize("status_code", (400, 404, 429, 500, 503))
@pytest.mark.parametrize("mock_release_id", range(3))
def test_edit_release_notes_fails(
    default_gh_client: Github, status_code: int, mock_release_id: int
):
    # Setup
    expected_num_requests = 1
    expected_http_method = "POST"
    expected_request_url = (
        "{api_url}/repos/{owner}/{repo_name}/releases/{release_id}".format(
            api_url=default_gh_client.api_url,
            owner=default_gh_client.owner,
            repo_name=default_gh_client.repo_name,
            release_id=mock_release_id,
        )
    )
    expected_request_body = {"body": RELEASE_NOTES}

    with requests_mock.Mocker(session=default_gh_client.session) as m:
        # mock the response
        m.register_uri(
            "POST",
            github_api_matcher,
            json={"id": mock_release_id},
            status_code=status_code,
        )

        # Execute method under test expecting an exception to be raised
        with pytest.raises(HTTPError):
            default_gh_client.edit_release_notes(mock_release_id, RELEASE_NOTES)

        # Evaluate (expected -> actual)
        assert m.called
        assert expected_num_requests == len(m.request_history)
        assert expected_http_method == m.last_request.method
        assert expected_request_url == m.last_request.url
        assert expected_request_body == m.last_request.json()


@pytest.mark.parametrize(
    "resp_payload, status_code, expected_result",
    [
        ({"id": 420, "status": "success"}, 200, 420),
        ({"error": "not found"}, 404, None),
        ({"error": "too many requests"}, 429, None),
        ({"error": "internal error"}, 500, None),
        ({"error": "temporarily unavailable"}, 503, None),
    ],
)
def test_get_release_id_by_tag(
    default_gh_client: Github,
    resp_payload: dict[str, int],
    status_code: int,
    expected_result: int | None,
):
    # Setup
    tag = "v1.0.0"
    expected_num_requests = 1
    expected_http_method = "GET"
    expected_request_url = (
        "{api_url}/repos/{owner}/{repo_name}/releases/tags/{tag}".format(
            api_url=default_gh_client.api_url,
            owner=default_gh_client.owner,
            repo_name=default_gh_client.repo_name,
            tag=tag,
        )
    )

    with requests_mock.Mocker(session=default_gh_client.session) as m:
        # mock the response
        m.register_uri(
            "GET", github_api_matcher, json=resp_payload, status_code=status_code
        )

        # Execute method under test
        rtn_val = default_gh_client.get_release_id_by_tag(tag)

        # Evaluate (expected -> actual)
        assert expected_result == rtn_val
        assert m.called
        assert expected_num_requests == len(m.request_history)
        assert expected_http_method == m.last_request.method
        assert expected_request_url == m.last_request.url


# Note - mocking as the logic for the create/update of a release
# is covered by testing above, no point re-testing.


@pytest.mark.parametrize("mock_release_id", range(3))
@pytest.mark.parametrize("prerelease", (True, False))
def test_create_or_update_release_when_create_succeeds(
    default_gh_client: Github,
    mock_release_id: int,
    prerelease: bool,
):
    tag = "v1.0.0"
    with mock.patch.object(
        default_gh_client,
        default_gh_client.create_release.__name__,
        return_value=mock_release_id,
    ) as mock_create_release, mock.patch.object(
        default_gh_client,
        default_gh_client.get_release_id_by_tag.__name__,
        return_value=mock_release_id,
    ) as mock_get_release_id_by_tag, mock.patch.object(
        default_gh_client,
        default_gh_client.edit_release_notes.__name__,
        return_value=mock_release_id,
    ) as mock_edit_release_notes:
        # Execute in mock environment
        result = default_gh_client.create_or_update_release(
            tag, RELEASE_NOTES, prerelease
        )

        # Evaluate (expected -> actual)
        assert mock_release_id == result
        mock_create_release.assert_called_once_with(tag, RELEASE_NOTES, prerelease)
        mock_get_release_id_by_tag.assert_not_called()
        mock_edit_release_notes.assert_not_called()


@pytest.mark.parametrize("mock_release_id", range(3))
@pytest.mark.parametrize("prerelease", (True, False))
def test_create_or_update_release_when_create_fails_and_update_succeeds(
    default_gh_client: Github,
    mock_release_id: int,
    prerelease: bool,
):
    tag = "v1.0.0"
    not_found = HTTPError("404 Not Found")
    not_found.response = Response()
    not_found.response.status_code = 404

    with mock.patch.object(
        default_gh_client,
        default_gh_client.create_release.__name__,
        side_effect=not_found,
    ) as mock_create_release, mock.patch.object(
        default_gh_client,
        default_gh_client.get_release_id_by_tag.__name__,
        return_value=mock_release_id,
    ) as mock_get_release_id_by_tag, mock.patch.object(
        default_gh_client,
        default_gh_client.edit_release_notes.__name__,
        return_value=mock_release_id,
    ) as mock_edit_release_notes:
        # Execute in mock environment
        result = default_gh_client.create_or_update_release(
            tag, RELEASE_NOTES, prerelease
        )

        # Evaluate (expected -> actual)
        assert mock_release_id == result
        mock_create_release.assert_called_once()
        mock_get_release_id_by_tag.assert_called_once_with(tag)
        mock_edit_release_notes.assert_called_once_with(mock_release_id, RELEASE_NOTES)


@pytest.mark.parametrize("prerelease", (True, False))
def test_create_or_update_release_when_create_fails_and_no_release_for_tag(
    default_gh_client: Github,
    prerelease: bool,
):
    tag = "v1.0.0"
    not_found = HTTPError("404 Not Found")
    not_found.response = Response()
    not_found.response.status_code = 404

    with mock.patch.object(
        default_gh_client,
        default_gh_client.create_release.__name__,
        side_effect=not_found,
    ) as mock_create_release, mock.patch.object(
        default_gh_client,
        default_gh_client.get_release_id_by_tag.__name__,
        return_value=None,
    ) as mock_get_release_id_by_tag, mock.patch.object(
        default_gh_client,
        default_gh_client.edit_release_notes.__name__,
        return_value=None,
    ) as mock_edit_release_notes:
        # Execute in mock environment expecting an exception to be raised
        with pytest.raises(ValueError):
            default_gh_client.create_or_update_release(tag, RELEASE_NOTES, prerelease)

        mock_create_release.assert_called_once()
        mock_get_release_id_by_tag.assert_called_once_with(tag)
        mock_edit_release_notes.assert_not_called()


def test_asset_upload_url(default_gh_client: Github):
    release_id = 1
    expected_num_requests = 1
    expected_http_method = "GET"
    expected_asset_upload_request_url = (
        "{api_url}/repos/{owner}/{repo}/releases/{release_id}".format(
            api_url=default_gh_client.api_url,
            owner=default_gh_client.owner,
            repo=default_gh_client.repo_name,
            release_id=release_id,
        )
    )
    mocked_upload_url = (
        "{upload_domain}/repos/{owner}/{repo}/releases/{release_id}/assets".format(
            upload_domain=github_upload_url,
            owner=default_gh_client.owner,
            repo=default_gh_client.repo_name,
            release_id=release_id,
        )
    )
    # '{?name,label}' are added by github.com at least, maybe custom too
    # https://docs.github.com/en/rest/releases/releases?apiVersion=2022-11-28#get-a-release
    resp_payload = {
        "upload_url": mocked_upload_url + "{?name,label}",
        "status": "success",
    }

    with requests_mock.Mocker(session=default_gh_client.session) as m:
        # mock the response
        m.register_uri("GET", github_api_matcher, json=resp_payload, status_code=200)

        # Execute method under test
        result = default_gh_client.asset_upload_url(release_id)

        # Evaluate (expected -> actual)
        assert mocked_upload_url == result
        assert m.called
        assert expected_num_requests == len(m.request_history)
        assert expected_http_method == m.last_request.method
        assert expected_asset_upload_request_url == m.last_request.url


@pytest.mark.parametrize("status_code", (200, 201))
@pytest.mark.parametrize("mock_release_id", range(3))
@pytest.mark.usefixtures(init_example_project.__name__)
def test_upload_release_asset_succeeds(
    default_gh_client: Github,
    example_changelog_md: Path,
    status_code: int,
    mock_release_id: int,
):
    # Setup
    label = "abc123"
    urlparams = {"name": example_changelog_md.name, "label": label}
    release_upload_url = (
        "{upload_domain}/repos/{owner}/{repo}/releases/{release_id}/assets".format(
            upload_domain=github_upload_url,
            owner=default_gh_client.owner,
            repo=default_gh_client.repo_name,
            release_id=mock_release_id,
        )
    )
    expected_num_requests = 2
    expected_retrieve_upload_url_method = "GET"
    expected_upload_http_method = "POST"
    expected_upload_url = "{url}?{params}".format(
        url=release_upload_url,
        params=urlencode(urlparams),
    )
    expected_changelog = example_changelog_md.read_bytes()
    json_get_up_url = {
        "status": "ok",
        "upload_url": release_upload_url + "{?name,label}",
    }

    with requests_mock.Mocker(session=default_gh_client.session) as m:
        # mock the responses
        m.register_uri(
            "POST",
            github_upload_matcher,
            json={"status": "ok"},
            status_code=status_code,
        )
        m.register_uri(
            "GET", github_api_matcher, json=json_get_up_url, status_code=status_code
        )

        # Execute method under test
        result = default_gh_client.upload_release_asset(
            release_id=mock_release_id,
            file=example_changelog_md.resolve(),
            label=label,
        )

        # Evaluate (expected -> actual)
        assert result is True
        assert m.called
        assert expected_num_requests == len(m.request_history)

        get_req, post_req = m.request_history

        assert expected_retrieve_upload_url_method == get_req.method
        assert expected_upload_http_method == post_req.method
        assert expected_upload_url == post_req.url
        assert expected_changelog == post_req.body


@pytest.mark.parametrize("status_code", (400, 404, 429, 500, 503))
@pytest.mark.parametrize("mock_release_id", range(3))
@pytest.mark.usefixtures(init_example_project.__name__)
def test_upload_release_asset_fails(
    default_gh_client: Github,
    example_changelog_md: Path,
    status_code: int,
    mock_release_id: int,
):
    # Setup
    label = "abc123"
    upload_url = "{up_url}/repos/{owner}/{repo_name}/releases/{release_id}".format(
        up_url=github_upload_url,
        owner=default_gh_client.owner,
        repo_name=default_gh_client.repo_name,
        release_id=mock_release_id,
    )
    json_get_up_url = {
        "status": "ok",
        "upload_url": upload_url,
    }

    with requests_mock.Mocker(session=default_gh_client.session) as m:
        # mock the responses
        m.register_uri(
            "POST",
            github_upload_matcher,
            json={"message": "error"},
            status_code=status_code,
        )
        m.register_uri("GET", github_api_matcher, json=json_get_up_url, status_code=200)

        # Execute method under test expecting an exception to be raised
        with pytest.raises(HTTPError):
            default_gh_client.upload_release_asset(
                release_id=mock_release_id,
                file=example_changelog_md.resolve(),
                label=label,
            )


# Note - mocking as the logic for uploading an asset
# is covered by testing above, no point re-testing.
def test_upload_dists_when_release_id_not_found(default_gh_client):
    tag = "v1.0.0"
    path = "doesn't matter"
    expected_num_uploads = 0

    # Set up mock environment
    with mock.patch.object(
        default_gh_client,
        default_gh_client.get_release_id_by_tag.__name__,
        return_value=None,
    ) as mock_get_release_id_by_tag, mock.patch.object(
        default_gh_client, default_gh_client.upload_release_asset.__name__
    ) as mock_upload_release_asset:
        # Execute method under test
        result = default_gh_client.upload_dists(tag, path)

        # Evaluate
        assert expected_num_uploads == result
        mock_get_release_id_by_tag.assert_called_once_with(tag=tag)
        mock_upload_release_asset.assert_not_called()


@pytest.mark.parametrize(
    "files, glob_pattern, upload_statuses, expected_num_uploads",
    [
        (["foo.zip", "bar.whl"], "*.zip", [True], 1),
        (["foo.whl", "foo.egg", "foo.tar.gz"], "foo.*", [True, True, True], 3),
        # What if not built?
        ([], "*", [], 0),
        # What if wrong directory/other stuff in output dir/subfolder?
        (["specialconfig.yaml", "something.whl", "desc.md"], "*.yaml", [True], 1),
        (["specialconfig.yaml", "something.whl", "desc.md"], "*.md", [True], 1),
    ],
)
def test_upload_dists_when_release_id_found(
    default_gh_client: Github,
    files: list[str],
    glob_pattern: str,
    upload_statuses: list[bool],
    expected_num_uploads: int,
):
    release_id = 420
    tag = "doesn't matter"
    matching_files = fnmatch.filter(files, glob_pattern)
    expected_files_uploaded = [mock.call(release_id, fn) for fn in matching_files]

    # Skip check as the files don't exist in filesystem
    mocked_isfile = mock.patch.object(os.path, "isfile", return_value=True)
    mocked_globber = mock.patch.object(glob, "glob", return_value=matching_files)

    # Set up mock environment
    with mocked_globber, mocked_isfile, mock.patch.object(
        default_gh_client,
        default_gh_client.get_release_id_by_tag.__name__,
        return_value=release_id,
    ) as mock_get_release_id_by_tag, mock.patch.object(
        default_gh_client,
        default_gh_client.upload_release_asset.__name__,
        side_effect=upload_statuses,
    ) as mock_upload_release_asset:
        # Execute method under test
        num_uploads = default_gh_client.upload_dists(tag, glob_pattern)

        # Evaluate (expected -> actual)
        assert expected_num_uploads == num_uploads
        mock_get_release_id_by_tag.assert_called_once_with(tag=tag)
        assert expected_files_uploaded == mock_upload_release_asset.call_args_list



from __future__ import annotations

import logging
import os
from dataclasses import dataclass
from enum import Enum
from pathlib import Path, PurePosixPath
from re import compile as regexp
from typing import TYPE_CHECKING, Any, Callable, Literal

from urllib3.util import Url

from semantic_release.const import PYPI_WEB_DOMAIN
from semantic_release.helpers import sort_numerically

if TYPE_CHECKING:  # pragma: no cover
    from jinja2 import Environment

    from semantic_release.changelog.release_history import Release, ReleaseHistory
    from semantic_release.hvcs._base import HvcsBase
    from semantic_release.version.version import Version


@dataclass
class ReleaseNotesContext:
    repo_name: str
    repo_owner: str
    hvcs_type: str
    version: Version
    release: Release
    mask_initial_release: bool
    license_name: str
    filters: tuple[Callable[..., Any], ...] = ()

    def bind_to_environment(self, env: Environment) -> Environment:
        env_globals = dict(
            filter(lambda k_v: k_v[0] != "filters", self.__dict__.items())
        )

        for g, v in env_globals.items():
            env.globals[g] = v

        for f in self.filters:
            env.filters[f.__name__] = f

        return env


class ChangelogMode(Enum):
    INIT = "init"
    UPDATE = "update"


@dataclass
class ChangelogContext:
    repo_name: str
    repo_owner: str
    hvcs_type: str
    history: ReleaseHistory
    changelog_mode: Literal["update", "init"]
    prev_changelog_file: str
    changelog_insertion_flag: str
    mask_initial_release: bool
    filters: tuple[Callable[..., Any], ...] = ()

    def bind_to_environment(self, env: Environment) -> Environment:
        env.globals["context"] = self
        env.globals["ctx"] = self
        for f in self.filters:
            env.filters[f.__name__] = f
        return env


def make_changelog_context(
    hvcs_client: HvcsBase,
    release_history: ReleaseHistory,
    mode: ChangelogMode,
    prev_changelog_file: Path,
    insertion_flag: str,
    mask_initial_release: bool,
) -> ChangelogContext:
    return ChangelogContext(
        repo_name=hvcs_client.repo_name,
        repo_owner=hvcs_client.owner,
        history=release_history,
        changelog_mode=mode.value,
        changelog_insertion_flag=insertion_flag,
        mask_initial_release=mask_initial_release,
        prev_changelog_file=str(prev_changelog_file),
        hvcs_type=hvcs_client.__class__.__name__.lower(),
        filters=(
            *hvcs_client.get_changelog_context_filters(),
            create_pypi_url,
            read_file,
            convert_md_to_rst,
            autofit_text_width,
            sort_numerically,
        ),
    )


def create_pypi_url(package_name: str, version: str = "") -> str:
    project_name = package_name.strip("/").strip()
    if not project_name:
        raise ValueError("package_name must not be empty!")
    return Url(
        scheme="https",
        host=PYPI_WEB_DOMAIN,
        path=str(PurePosixPath("project", project_name, version.strip("/").strip())),
    ).url.rstrip("/")


def read_file(filepath: str) -> str:
    try:
        if not filepath:
            raise FileNotFoundError("No file path provided")  # noqa: TRY301

        with Path(filepath).open(newline=os.linesep) as rfd:
            return rfd.read()

    except FileNotFoundError as err:
        logging.warning(str(err))
        return ""


def convert_md_to_rst(md_content: str) -> str:
    rst_content = md_content
    replacements = {
        # Replace markdown doubleunder bold with rst bold
        "bold-inline": (regexp(r"(?<=\s)__(.+?)__(?=\s|$)"), r"**\1**"),
        # Replace markdown italics with rst italics
        "italic-inline": (regexp(r"(?<=\s)_([^_].+?[^_])_(?=\s|$)"), r"*\1*"),
        # Replace markdown bullets with rst bullets
        "bullets": (regexp(r"^(\s*)-(\s)"), r"\1*\2"),
        # Replace markdown inline raw content with rst inline raw content
        "raw-inline": (regexp(r"(?<=\s)(`[^`]+`)(?![`_])"), r"`\1`"),
        # Replace markdown inline link with rst inline link
        "link-inline": (
            regexp(r"(?<=\s)\[([^\]]+)\]\(([^)]+)\)(?=\s|$)"),
            r"`\1 <\2>`_",
        ),
    }

    for pattern, replacement in replacements.values():
        rst_content = pattern.sub(replacement, rst_content)

    return rst_content


def autofit_text_width(text: str, maxwidth: int = 100, indent_size: int = 0) -> str:
    """Format the description text to fit within a specified width"""
    input_text = text.strip()

    if len(input_text) <= maxwidth:
        # If the text is already within the maxwidth, return immediately
        return input_text

    indent = " " * indent_size
    formatted_description = []

    # Re-format text to fit within the maxwidth
    for paragraph in input_text.split("\n\n"):
        formatted_paragraph = []

        # Split the paragraph into words with no empty strings
        words = list(
            filter(
                None, paragraph.replace("\r", "").replace("\n", " ").strip().split(" ")
            )
        )

        # Initialize the line for each paragraph
        line = words[0]
        next_line = ""

        for word in words[1:]:
            # Check if the current line + the next word (and a space) will fit within the maxwidth
            # If it does, then update the current line
            next_line = f"{line} {word}"
            if len(next_line) <= maxwidth:
                line = next_line
                continue

            # Add the current line to the paragraph and start a new line
            formatted_paragraph.append(line)
            line = f"{indent}{word}"

        # Store the last line in the paragraph since it hasn't reached the maxwidth yet
        formatted_paragraph.append(line)

        #
        formatted_description.append(str.join("\n", formatted_paragraph))

    # Print the formatted description
    return str.join("\n\n", formatted_description).strip()
